"""Provides Bokeh based utilities for inspecting fits and fit quality."""
import numpy as np
from bokeh import events

import xarray as xr
from arpes.plotting.interactive_utils import BokehInteractiveTool, CursorTool

__all__ = ("FitCheckTool",)


class FitCheckTool(BokehInteractiveTool, CursorTool):
    """Interactive verification of fit quality for broadcast fits."""

    auto_zero_nans = False
    auto_rebin = False

    def __init__(self, **kwargs):
        """Loads marginal sizes and configures initial application settings."""
        super().__init__()

        self.load_settings(**kwargs)
        self.app_main_size = self.settings.get("main_width", 600)
        self.app_marginal_size = self.settings.get("marginal_width", 300)
        self.selected_data = "data"
        self.use_dataset = True
        self.remove_outliers = True
        self.outlier_clip = 1

    def tool_handler(self, doc):
        """Defines the application widgets and UI interactions."""
        from bokeh.layouts import row, column, Spacer
        from bokeh.models.mappers import LinearColorMapper
        from bokeh.models import widgets
        from bokeh.models.widgets.markups import Div
        from bokeh.plotting import figure

        self.arr = self.arr.copy(deep=True)

        if not isinstance(self.arr, xr.Dataset):
            self.use_dataset = False

        residual = None
        if self.use_dataset:
            raw_data = self.arr.data
            raw_data.values[np.isnan(raw_data.values)] = 0
            fit_results = self.arr.results
            residual = self.arr.residual
            residual.values[np.isnan(residual.values)] = 0
        else:
            raw_data = self.arr.attrs["original_data"]
            fit_results = self.arr

        fit_direction = [d for d in raw_data.dims if d not in fit_results.dims]
        fit_direction = fit_direction[0]

        two_dimensional = False
        if len(raw_data.dims) != 2:
            two_dimensional = True
            x_coords, y_coords = (
                fit_results.coords[fit_results.dims[0]],
                fit_results.coords[fit_results.dims[1]],
            )
            z_coords = raw_data.coords[fit_direction]
        else:
            x_coords, y_coords = (
                raw_data.coords[raw_data.dims[0]],
                raw_data.coords[raw_data.dims[1]],
            )

        if two_dimensional:
            self.settings["palette"] = "coolwarm"
        default_palette = self.default_palette

        self.app_context.update(
            {
                "data": raw_data,
                "fits": fit_results,
                "residual": residual,
                "original": self.arr,
                "data_range": {
                    "x": (np.min(x_coords.values), np.max(x_coords.values)),
                    "y": (np.min(y_coords.values), np.max(y_coords.values)),
                },
            }
        )
        if two_dimensional:
            self.app_context["data_range"]["z"] = (np.min(z_coords.values), np.max(z_coords.values))

        figures, plots, app_widgets = (
            self.app_context["figures"],
            self.app_context["plots"],
            self.app_context["widgets"],
        )

        self.cursor_dims = raw_data.dims
        if two_dimensional:
            self.cursor = [
                np.mean(self.data_range["x"]),
                np.mean(self.data_range["y"]),
                np.mean(self.data_range["z"]),
            ]
        else:
            self.cursor = [np.mean(self.data_range["x"]), np.mean(self.data_range["y"])]

        app_widgets["fit_info_div"] = Div(text="")

        self.app_context["color_maps"]["main"] = LinearColorMapper(
            default_palette,
            low=np.min(raw_data.values),
            high=np.max(raw_data.values),
            nan_color="black",
        )

        main_tools = ["wheel_zoom", "tap", "reset", "save"]
        main_title = "Fit Inspection Tool: WARNING Unidentified"

        try:
            main_title = "Fit Inspection Tool: {}".format(raw_data.S.label[:60])
        except:
            pass

        figures["main"] = figure(
            tools=main_tools,
            plot_width=self.app_main_size,
            plot_height=self.app_main_size,
            min_border=10,
            min_border_left=50,
            toolbar_location="left",
            x_axis_location="below",
            y_axis_location="right",
            title=main_title,
            x_range=self.data_range["x"],
            y_range=self.app_context["data_range"]["y"],
        )
        figures["main"].xaxis.axis_label = raw_data.dims[0]
        figures["main"].yaxis.axis_label = raw_data.dims[1]
        figures["main"].toolbar.logo = None
        figures["main"].background_fill_color = "#fafafa"

        data_for_main = raw_data
        if two_dimensional:
            data_for_main = data_for_main.sel(
                **dict([[fit_direction, self.cursor[2]]]), method="nearest"
            )
        plots["main"] = figures["main"].image(
            [data_for_main.values.T],
            x=self.app_context["data_range"]["x"][0],
            y=self.app_context["data_range"]["y"][0],
            dw=self.app_context["data_range"]["x"][1] - self.app_context["data_range"]["x"][0],
            dh=self.app_context["data_range"]["y"][1] - self.app_context["data_range"]["y"][0],
            color_mapper=self.app_context["color_maps"]["main"],
        )

        band_centers = [b.center for b in fit_results.F.bands.values()]
        bands_xs = [b.coords[b.dims[0]].values for b in band_centers]
        bands_ys = [b.values for b in band_centers]
        if fit_results.dims[0] == raw_data.dims[1]:
            bands_ys, bands_xs = bands_xs, bands_ys
        plots["band_locations"] = figures["main"].multi_line(
            xs=bands_xs, ys=bands_ys, line_color="white", line_width=1, line_dash="dashed"
        )

        # add cursor lines
        cursor_lines = self.add_cursor_lines(figures["main"])

        # marginals
        if not two_dimensional:
            figures["bottom"] = figure(
                plot_width=self.app_main_size,
                plot_height=self.app_marginal_size,
                min_border=10,
                title=None,
                x_range=figures["main"].x_range,
                x_axis_location="above",
                toolbar_location=None,
                tools=[],
            )
        else:
            figures["bottom"] = Spacer(width=self.app_main_size, height=self.app_marginal_size)

        right_y_range = figures["main"].y_range
        if two_dimensional:
            right_y_range = self.data_range["z"]

        figures["right"] = figure(
            plot_width=self.app_marginal_size,
            plot_height=self.app_main_size,
            min_border=10,
            title=None,
            y_range=right_y_range,
            y_axis_location="left",
            toolbar_location=None,
            tools=[],
        )

        marginal_line_width = 2
        if not two_dimensional:
            bottom_data = raw_data.sel(
                **dict([[raw_data.dims[1], self.cursor[1]]]), method="nearest"
            )
            right_data = raw_data.sel(
                **dict([[raw_data.dims[0], self.cursor[0]]]), method="nearest"
            )

            plots["bottom"] = figures["bottom"].line(
                x=bottom_data.coords[raw_data.dims[0]].values,
                y=bottom_data.values,
                line_width=marginal_line_width,
            )
            plots["bottom_residual"] = figures["bottom"].line(
                x=[], y=[], line_color="red", line_width=marginal_line_width
            )
            plots["bottom_fit"] = figures["bottom"].line(
                x=[], y=[], line_color="blue", line_width=marginal_line_width, line_dash="dashed"
            )
            plots["bottom_init_fit"] = figures["bottom"].line(
                x=[], y=[], line_color="green", line_width=marginal_line_width, line_dash="dotted"
            )

            plots["right"] = figures["right"].line(
                y=right_data.coords[raw_data.dims[1]].values,
                x=right_data.values,
                line_width=marginal_line_width,
            )
            plots["right_residual"] = figures["right"].line(
                x=[], y=[], line_color="red", line_width=marginal_line_width
            )
            plots["right_fit"] = figures["right"].line(
                x=[], y=[], line_color="blue", line_width=marginal_line_width, line_dash="dashed"
            )
            plots["right_init_fit"] = figures["right"].line(
                x=[], y=[], line_color="green", line_width=marginal_line_width, line_dash="dotted"
            )
        else:
            right_data = raw_data.sel(
                **{k: v for k, v in self.cursor_dict.items() if k != fit_direction},
                method="nearest"
            )
            plots["right"] = figures["right"].line(
                y=right_data.coords[right_data.dims[0]].values,
                x=right_data.values,
                line_width=marginal_line_width,
            )
            plots["right_residual"] = figures["right"].line(
                x=[], y=[], line_color="red", line_width=marginal_line_width
            )
            plots["right_fit"] = figures["right"].line(
                x=[], y=[], line_color="blue", line_width=marginal_line_width, line_dash="dashed"
            )
            plots["right_init_fit"] = figures["right"].line(
                x=[], y=[], line_color="green", line_width=marginal_line_width, line_dash="dotted"
            )

        def on_change_main_view(event):
            data_source = event.item
            self.selected_data = data_source
            data = None
            if data_source == "data":
                data = raw_data.sel(
                    **{k: v for k, v in self.cursor_dict.items() if k == fit_direction},
                    method="nearest"
                )
            elif data_source == "residual":
                data = residual.sel(
                    **{k: v for k, v in self.cursor_dict.items() if k == fit_direction},
                    method="nearest"
                )
            elif two_dimensional:
                data = fit_results.F.s(data_source)
                data.values[np.isnan(data.values)] = 0

            if data is not None:
                if self.remove_outliers:
                    data = data.G.clean_outliers(clip=self.outlier_clip)

                plots["main"].data_source.data = {
                    "image": [data.values.T],
                }
                update_main_colormap(None, None, main_color_range_slider.value)

        def update_fit_display():
            target = "right"
            if fit_results.dims[0] == raw_data.dims[1]:
                target = "bottom"

            if two_dimensional:
                target = "right"
                current_fit = fit_results.sel(
                    **{k: v for k, v in self.cursor_dict.items() if k != fit_direction},
                    method="nearest"
                ).item()
                coord_vals = raw_data.coords[fit_direction].values
            else:
                current_fit = fit_results.sel(
                    **dict([[fit_results.dims[0], self.cursor[0 if target == "right" else 1]]]),
                    method="nearest"
                ).item()
                coord_vals = raw_data.coords[raw_data.dims[0 if target == "bottom" else 1]].values

            if current_fit is not None:
                app_widgets["fit_info_div"].text = current_fit._repr_html_(
                    short=True
                )  # pylint: disable=protected-access
            else:
                app_widgets["fit_info_div"].text = "No fit here."
                plots["{}_residual".format(target)].data_source.data = {
                    "x": [],
                    "y": [],
                }
                plots["{}_fit".format(target)].data_source.data = {
                    "x": [],
                    "y": [],
                }
                plots["{}_init_fit".format(target)].data_source.data = {
                    "x": [],
                    "y": [],
                }
                return

            if target == "bottom":
                residual_x = coord_vals
                residual_y = current_fit.residual
                init_fit_x = coord_vals
                init_fit_y = current_fit.init_fit
                fit_x = coord_vals
                fit_y = current_fit.best_fit
            else:
                residual_y = coord_vals
                residual_x = current_fit.residual
                init_fit_y = coord_vals
                init_fit_x = current_fit.init_fit
                fit_y = coord_vals
                fit_x = current_fit.best_fit

            plots["{}_residual".format(target)].data_source.data = {
                "x": residual_x,
                "y": residual_y,
            }
            plots["{}_fit".format(target)].data_source.data = {
                "x": fit_x,
                "y": fit_y,
            }
            plots["{}_init_fit".format(target)].data_source.data = {
                "x": init_fit_x,
                "y": init_fit_y,
            }

        def click_right_marginal(event):
            self.cursor = [self.cursor[0], self.cursor[1], event.y]
            on_change_main_view(None, None, self.selected_data)

        def click_main_image(event):
            if two_dimensional:
                self.cursor = [event.x, event.y, self.cursor[2]]
            else:
                self.cursor = [event.x, event.y]

            if not two_dimensional:
                right_marginal_data = raw_data.sel(
                    **dict([[raw_data.dims[0], self.cursor[0]]]), method="nearest"
                )
                bottom_marginal_data = raw_data.sel(
                    **dict([[raw_data.dims[1], self.cursor[1]]]), method="nearest"
                )
                plots["bottom"].data_source.data = {
                    "x": bottom_marginal_data.coords[raw_data.dims[0]].values,
                    "y": bottom_marginal_data.values,
                }
            else:
                right_marginal_data = raw_data.sel(
                    **{k: v for k, v in self.cursor_dict.items() if k != fit_direction},
                    method="nearest"
                )

            plots["right"].data_source.data = {
                "y": right_marginal_data.coords[right_marginal_data.dims[0]].values,
                "x": right_marginal_data.values,
            }

            update_fit_display()

        def on_change_outlier_clip(attr, old, new):
            self.outlier_clip = new
            on_change_main_view(None, None, self.selected_data)

        def set_remove_outliers(should_remove_outliers):
            if self.remove_outliers != should_remove_outliers:
                self.remove_outliers = should_remove_outliers

                on_change_main_view(None, None, self.selected_data)

        update_main_colormap = self.update_colormap_for("main")
        MAIN_CONTENT_OPTIONS = [
            ("Residual", "residual"),
            ("Data", "data"),
        ]

        if two_dimensional:
            available_parameters = fit_results.F.parameter_names

            for param_name in available_parameters:
                MAIN_CONTENT_OPTIONS.append(
                    (
                        param_name,
                        param_name,
                    )
                )

        remove_outliers_toggle = widgets.Toggle(
            label="Remove Outliers", button_type="primary", active=self.remove_outliers
        )
        remove_outliers_toggle.on_click(set_remove_outliers)

        outlier_clip_slider = widgets.Slider(
            title="Clip", start=0, end=10, value=self.outlier_clip, callback_throttle=150, step=0.2
        )
        outlier_clip_slider.on_change("value", on_change_outlier_clip)

        main_content_select = widgets.Dropdown(
            label="Main Content", button_type="primary", menu=MAIN_CONTENT_OPTIONS
        )
        main_content_select.on_click(on_change_main_view)

        # Widgety things
        main_color_range_slider = widgets.RangeSlider(
            start=0,
            end=100,
            value=(
                0,
                100,
            ),
            title="Color Range",
        )

        # Attach callbacks
        main_color_range_slider.on_change("value", update_main_colormap)
        figures["main"].on_event(events.Tap, click_main_image)
        if two_dimensional:
            figures["right"].on_event(events.Tap, click_right_marginal)

        layout = row(
            column(figures["main"], figures.get("bottom")),
            column(figures["right"], app_widgets["fit_info_div"]),
            column(
                column(
                    *[
                        widget
                        for widget in [
                            self._cursor_info,
                            main_color_range_slider,
                            main_content_select,
                            remove_outliers_toggle if two_dimensional else None,
                            outlier_clip_slider if two_dimensional else None,
                        ]
                        if widget is not None
                    ]
                ),
            ),
        )

        update_fit_display()

        doc.add_root(layout)
        doc.title = "Band Tool"



"""Utilities for inspecting fit results by hand by plotting them individually."""
import matplotlib.pyplot as plt
import numpy as np

from arpes.plotting.utils import simple_ax_grid

__all__ = (
    "plot_fit",
    "plot_fits",
)


def plot_fit(model_result, ax=None):
    """Performs a straightforward plot of the data, residual, and fit to an axis."""
    if ax is None:
        fig, ax = plt.subplots()

    x = model_result.userkws[model_result.model.independent_vars[0]]
    ax2 = ax.twinx()
    ax2.grid(False)
    ax2.axhline(0, color="green", linestyle="--", alpha=0.5)

    ax.scatter(x, model_result.data, s=10, edgecolors="blue", marker="s", c="white", linewidth=1.5)
    ax.plot(x, model_result.best_fit, color="red", linewidth=1.5)

    ax2.scatter(
        x,
        model_result.residual,
        edgecolors="green",
        alpha=0.5,
        s=12,
        marker="s",
        c="white",
        linewidth=1.5,
    )
    ylim = np.max(np.abs(np.asarray(ax2.get_ylim()))) * 1.5
    ax2.set_ylim([-ylim, ylim])
    ax.set_xlim([np.min(x), np.max(x)])


def plot_fits(model_results, ax=None):
    """Plots several fits onto a grid of axes."""
    n_results = len(model_results)
    if ax is None:
        fig, ax, ax_extra = simple_ax_grid(n_results, sharex="col", sharey="row")

    for axi, model_result in zip(ax, model_results):
        plot_fit(model_result, ax=axi)



"""For now we monkeypatch lmfit to make it easier to work with in Jupyter.

We should consider forking or providing a pull at a later date after this settles down.

The end goal here is to allow pleasing and functional representations of curve fitting sessions
performed in Jupyter, so that they can be rapidly understood, and screencapped for simple purposes,
like including in group meeting notes.
"""

# pylint: disable=protected-access

import numpy as np
from lmfit import model


def repr_multiline_ModelResult(self, **kwargs):
    """Provides a text-based multiline representation used in Qt based interactive tools."""
    template = "ModelResult\n  Converged: {success}\n  Components:\n {formatted_components}\n  Parameters:\n{parameters}"

    return template.format(
        success=self.success,
        formatted_components="\n".join(
            [(" " * 4) + c._repr_multiline_text_() for c in self.components]
        ),
        parameters="\n".join(
            f"    {l}" for l in self.params._repr_multiline_text_(**kwargs).split("\n")
        ),
    )


def repr_html_ModelResult(self, **kwargs):
    """Provides a better Jupyter representation of an `lmfit.ModelResult` instance."""
    template = """
        <div>
            <span><strong>Converged: </strong>{success}</span>
            <div>{formatted_components}<div>
            <div>{parameters}</div>
        </div>
        """
    return template.format(
        success=self.success,
        formatted_components="".join(
            "<div>{}</div>".format(c._repr_html_()) for c in self.components
        ),
        parameters=self.params._repr_html_(**kwargs),
    )


def repr_html_Model(self):
    """Better Jupyter representation of `lmfit.Model` instances."""
    template = """
    <div>
    <strong>{name}</strong>
    </div>
    """
    return template.format(name=self.name)


def repr_multiline_Model(self, **kwargs):
    """Provides a text-based multiline representation used in Qt based interactive tools."""
    return self.name


ALL_PARAMETER_ATTRIBUTES = ["name", "value", "min", "max", "stderr", "vary", "expr", "brute_step"]
SKIP_ON_SHORT = {"min", "max", "vary", "expr", "brute_step"}


def repr_html_Parameters(self, short=False):
    """HTML representation for `lmfit.Parameters` instances."""
    keys = sorted(list(self.keys()))
    template = """
    <table>
      <thead>
          <tr>
            {cols}
          </tr>
      </thead>
      <tbody>
        {rows}
      </tbody>
    </table>
    """
    return template.format(
        cols="".join(
            "<th>{}</th>".format(c)
            for c in ALL_PARAMETER_ATTRIBUTES
            if not short or c not in SKIP_ON_SHORT
        ),
        rows="".join(self[p].to_table_row(short=short) for p in keys),
    )


def repr_multiline_Parameters(self, short=False):
    """Provides a text-based multiline representation used in Qt based interactive tools."""
    return "\n".join(self[k]._repr_multiline_text_(short=short) for k in self.keys())


def repr_html_Parameter(self, short=False):
    """HTML representation for `lmfit.Parameter` instances."""
    if short:
        return """
            <tr>
                <th>{name}</th>
                <th>{value:.3f}</th>
                <th>{stderr:.3f}</th>
            </tr>
            """.format(
            name=self.name,
            value=self.value,
            stderr=self.stderr,
        )

    template = """
            <tr>
                <th>{name}</th>
                <th>{value:.3f}</th>
                <th>{min:.3f}</th>
                <th>{max:.3f}</th>
                <th>{stderr:.3f}</th>
                <th>{vary}</th>
                <th>{expr}</th>
                <th>{brute_step}</th>
            </tr>
            """
    return template.format(
        name=self.name,
        value=self.value,
        min=self.min,
        max=self.max,
        stderr=self.stderr or np.inf,
        vary=self.vary,
        expr=self.expr or "",
        brute_step=self.brute_step or "",
    )


def repr_multiline_Parameter(self: model.Parameter, short=False):
    """Provides a text-based multiline representation used in Qt based interactive tools."""
    template = "{name}:\n{contents}"

    get_attrs = [a for a in ALL_PARAMETER_ATTRIBUTES if not short or a not in SKIP_ON_SHORT]

    def format_attr(value) -> str:
        if isinstance(value, float):
            return f"{value:.3f}"

        return str(value)

    return template.format(
        name=self.name,
        contents="\n".join(
            f"  {attr_name}: {format_attr(getattr(self, attr_name))}"
            for attr_name in get_attrs
            if attr_name != "name"
        ),
    )


model.Model._repr_html_ = repr_html_Model
model.Model._repr_multiline_text_ = repr_multiline_Model
model.Parameters._repr_html_ = repr_html_Parameters
model.Parameters._repr_multiline_text_ = repr_multiline_Parameters
model.ModelResult._repr_html_ = repr_html_ModelResult
model.ModelResult._repr_multiline_text_ = repr_multiline_ModelResult

model.Parameter.to_table_row = repr_html_Parameter
model.Parameter._repr_multiline_text_ = repr_multiline_Parameter
# model.Parameter._repr_html_ = repr_html_Parameter

# we don't export anything, just monkey-patch
__all__ = tuple()



"""Provides data loading for the Lanzara group SPEEM."""

import copy

import pickle
import numpy as np
from scipy.interpolate import griddata

import xarray as xr
from arpes.endstations import EndstationBase
from arpes.provenance import provenance_from_file

__all__ = "SPEEMEndstation"

NS_PER_PIXEL = 4.576e-3
NOMINAL_TIMING_OFFSET = 921  # ns

# def pixels_to_time(
#     data: list,
#     pulse_delay: int = 1500,
#     time_per_pixel: float = 0.05,
#     nominal_delay: int = 1171,
# ):
#     return [pulse_delay - nominal_delay - value * time_per_pixel for value in data]

# def load_strict_range(
#     root: Path, x_range: tuple = None, y_range: tuple = None, t_range: tuple = None
# ):
#     with open(root / "raw_daq.pickle", "rb") as f:
#         raw_data = np.concatenate(pickle.load(f)["detector-frame-data"].data)

#     histogram: np.ndarray
#     bins = [
#         this_range[1] - this_range[0] if this_range is not None else 500
#         for this_range in [x_range, y_range, t_range]
#     ]
#     histogram, bins = np.histogramdd(
#         raw_data, bins=bins, range=[x_range, y_range, t_range]
#     )

#     # def midpoints(array: np.ndarray):
#     #     return [(a + b) / 2 for a, b in zip(array[:-1], array[1:])]
#     # coords = midpoints(bins)

#     coords = [bin[:-1] for bin in bins]

#     xr = DataArray(histogram, coords=coords, dims=["x", "y", "t"],)

#     return xr


class SPEEMEndstation(EndstationBase):
    """Provides data loading for the Lanzara group SPEEM."""

    PRINCIPAL_NAME = "ALG-SPEEM"
    ALIASES = ["SPEEM", "ToF", "PEEM-ToF"]
    _TOLERATED_EXTENSIONS = {".pickle"}

    def coordinate_conversion(
        self, count_list: np.ndarray, conversion_table: np.ndarray
    ) -> np.ndarray:
        """
        Converts x, y, t detector coordinates to photoemission angle/position and
        energy.
        """
        radius = np.hypot(count_list[:, 0], count_list[:, 1])
        theta = np.arctan2(count_list[:, 1], count_list[:, 0])

        method = "cubic"
        kinetic_energy = griddata(
            conversion_table[:, 0:2],
            conversion_table[:, 2],
            count_list[:, 2],
            method=method,
        )
        radial_angle = griddata(
            conversion_table[:, 0:2], conversion_table[:, 3], radius, method=method
        )

        phi = np.multiply(radial_angle, np.cos(theta))
        psi = np.multiply(radial_angle, np.sin(theta))

        return np.stack((phi, psi, kinetic_energy), axis=1)

    def load(self, scan_desc: dict, conversion_table: np.ndarray, **kwargs):
        """
        Loads a pickle file from the SPEEM DAQ.

        Args:
            scan_desc: Dictionary with extra information to attach to the
            xarray.Dataset, must contain the location of the file

        Returns:
            The loaded spectrum.
        """

        metadata = copy.deepcopy(scan_desc)

        data_loc = metadata["file"]

        dataset_contents = dict()

        with open(data_loc, "rb") as f:
            raw_counts = pickle.load(f)["detector-frame-data"].data
        # concatenate = kwargs.get("concatenate", True)
        # raw_counts = [np.concatenate(raw_counts)] if concatenate else raw_counts
        # dataset_contents["raw_counts"] = raw_counts

        raw_counts: np.ndarray = np.concatenate(raw_counts)
        converted_counts = self.coordinate_conversion(raw_counts, conversion_table)

        n_bins = kwargs.get("bins", 100)
        histogram, bins = np.histogramdd(converted_counts, bins=n_bins)
        coords = {"phi": bins[0], "psi": bins[1], "KE": bins[2]}
        dataset_contents["raw"] = xr.DataArray(
            histogram, coords=coords, dims=tuple(coords.keys())
        )

        # timing_delay = (
        #     kwargs.get("timing_delay", NOMINAL_TIMING_OFFSET) - NOMINAL_TIMING_OFFSET
        # )  # ns
        # energy_offset = kwargs.get("energy_offset", 0)  # eV
        # energy_scale = kwargs.get("energy_scale", 16)  # ns/eV

        # datasets = []
        # for frame in raw_counts:
        #     pruned_frame = self.prune_data(frame)
        #     histogram, bins = np.histogramdd(pruned_frame, bins=n_bins)
        #     coords = {"phi": bins[0][1:], "psi": bins[1][1:], "KE": bins[2][1:]}

        #     dataset_contents["raw"] = xr.DataArray(
        #         histogram,
        #         coords=coords,
        #         # coords={
        #         #     "x": np.linspace(-12, 12, bins),
        #         #     "y": np.linspace(-12, 12, bins),
        #         #     "Eb": np.linspace(
        #         #         (timing_delay - 4095 * NS_PER_PIXEL) / energy_scale + energy_offset,
        #         #         timing_delay / energy_scale + energy_offset,
        #         #         bins,
        #         #     ),
        #         # },
        #         dims=tuple(coords.keys()),
        #         # attrs=f["/PRIMARY"].attrs.items(),
        #     )

        #     datasets.append(dataset_contents)

        provenance_from_file(
            dataset_contents["raw"],
            data_loc,
            {
                "what": "Loaded SPEEM dataset.",
                "by": "SPEEM.load",
            },
        )

        # if concatenate:
        #     return xr.Dataset(datasets[0], attrs=metadata)
        return xr.Dataset(dataset_contents, attrs=metadata)

    @staticmethod
    def prune_data(data: np.ndarray):
        pruned = []
        for count in data:
            if 4095 not in count and 0 not in count:
                pruned.append(count)

        return np.array(pruned)



"""Utilities for plotting parameter data out of bulk fits."""
import matplotlib.pyplot as plt
from arpes.plotting.utils import latex_escape

import xarray as xr
from arpes.provenance import save_plot_provenance

__all__ = ("plot_parameter",)


@save_plot_provenance
def plot_parameter(
    fit_data: xr.DataArray,
    param_name: str,
    ax=None,
    fillstyle="none",
    shift=0,
    x_shift=0,
    markersize=8,
    title=None,
    out=None,
    two_sigma=False,
    **kwargs
):
    """Makes a simple scatter plot of a parameter from an `broadcast_fit` result."""
    fig = None

    if ax is None:
        fig, ax = plt.subplots(figsize=kwargs.get("figsize", (7, 5)))

    ds = fit_data.F.param_as_dataset(param_name)
    x_name = ds.value.dims[0]
    x = ds.coords[x_name].values

    color = kwargs.get("color")
    e_width = None
    l_width = None
    if two_sigma:
        line, markers, lines = ax.errorbar(
            x + x_shift,
            ds.value.values + shift,
            yerr=2 * ds.error.values,
            fmt="",
            elinewidth=1,
            linewidth=0,
            c=color,
        )
        color = lines[0].get_color()[0]
        e_width = 2
        l_width = 0

    ax.errorbar(
        x + x_shift,
        ds.value.values + shift,
        yerr=ds.error.values,
        fmt="s",
        color=color,
        elinewidth=e_width,
        linewidth=l_width,
        markeredgewidth=e_width or 2,
        fillstyle=fillstyle,
        markersize=markersize,
    )

    ax.set_xlabel(latex_escape(x_name))
    ax.set_ylabel(latex_escape(param_name))
    return ax



"""Some bespoke exceptions that can be used in control sequences.

Over builtins, these provide more information to the user. I (Conrad) prefer to use warnings for
the latter purpose, but there are reasons to throw these errors in a variety of circumstances.
"""


class AnalysisError(Exception):
    """Base class to indicate that something scientific went wrong.

    Example:
        A bad fit from scipy.optimize in an internal function or analysis
        routine that could not be handled by the user.
    """



"""Includes multi-peak model definitions."""

import lmfit as lf

from .x_model_mixin import XModelMixin
from .functional_forms import gaussian, affine_bkg, lorentzian, twolorentzian
from lmfit.models import update_param_vals

__all__ = ["TwoGaussianModel", "TwoLorModel"]


class TwoGaussianModel(XModelMixin):
    """A model for two gaussian functions with a linear background."""

    @staticmethod
    def twogaussian(
        x, center=0, t_center=0, width=1, t_width=1, amp=1, t_amp=1, lin_bkg=0, const_bkg=0
    ):
        """Two gaussians and an affine background."""
        return (
            gaussian(x, center, width, amp)
            + gaussian(x, t_center, t_width, t_amp)
            + affine_bkg(x, lin_bkg, const_bkg)
        )

    def __init__(self, independent_vars=("x",), prefix="", missing="raise", name=None, **kwargs):
        """Sets physical constraints for peak width and other parameters."""
        kwargs.update({"prefix": prefix, "missing": missing, "independent_vars": independent_vars})
        super().__init__(self.twogaussian, **kwargs)

        self.set_param_hint("amp", min=0.0)
        self.set_param_hint("width", min=0)
        self.set_param_hint("t_amp", min=0.0)
        self.set_param_hint("t_width", min=0)
        self.set_param_hint("lin_bkg", min=-10, max=10)
        self.set_param_hint("const_bkg", min=-50, max=50)

    def guess(self, data, x=None, **kwargs):
        """Very simple heuristics for peak location."""
        pars = self.make_params()

        pars["%scenter" % self.prefix].set(value=0)
        pars["%st_center" % self.prefix].set(value=0)
        pars["%slin_bkg" % self.prefix].set(value=0)
        pars["%sconst_bkg" % self.prefix].set(value=data.min())
        pars["%swidth" % self.prefix].set(0.02)  # TODO we can do better than this
        pars["%st_width" % self.prefix].set(0.02)
        pars["%samp" % self.prefix].set(value=data.mean() - data.min())
        pars["%st_amp" % self.prefix].set(value=data.mean() - data.min())

        return update_param_vals(pars, self.prefix, **kwargs)

    __init__.doc = lf.models.COMMON_INIT_DOC
    guess.__doc__ = lf.models.COMMON_GUESS_DOC


class TwoLorModel(XModelMixin):
    """A model for two gaussian functions with a linear background."""

    def __init__(self, independent_vars=("x",), prefix="", missing="raise", name=None, **kwargs):
        """Sets physical constraints for peak width and other parameters."""
        kwargs.update({"prefix": prefix, "missing": missing, "independent_vars": independent_vars})
        super().__init__(twolorentzian, **kwargs)

        self.set_param_hint("amp", min=0.0)
        self.set_param_hint("gamma", min=0)
        self.set_param_hint("t_amp", min=0.0)
        self.set_param_hint("t_gamma", min=0)
        self.set_param_hint("lin_bkg", min=-10, max=10)
        self.set_param_hint("const_bkg", min=-50, max=50)

    def guess(self, data, x=None, **kwargs):
        """Very simple heuristics for peak location."""
        pars = self.make_params()

        pars["%scenter" % self.prefix].set(value=0)
        pars["%st_center" % self.prefix].set(value=0)
        pars["%slin_bkg" % self.prefix].set(value=0)
        pars["%sconst_bkg" % self.prefix].set(value=data.min())
        pars["%sgamma" % self.prefix].set(0.02)  # TODO we can do better than this
        pars["%st_gamma" % self.prefix].set(0.02)
        pars["%samp" % self.prefix].set(value=data.mean() - data.min())
        pars["%st_amp" % self.prefix].set(value=data.mean() - data.min())

        return update_param_vals(pars, self.prefix, **kwargs)

    __init__.doc = lf.models.COMMON_INIT_DOC
    guess.__doc__ = lf.models.COMMON_GUESS_DOC



"""X-ray photoelectron spectroscopy related analysis.

Primarily, curve fitting and peak-finding utilities for XPS.
"""

import numpy as np

from arpes.analysis.general import rebin
from arpes.analysis.savitzky_golay import savitzky_golay
from arpes.typing import DataType
from arpes.utilities import normalize_to_spectrum

__all__ = ("approximate_core_levels",)


def local_minima(a: np.ndarray, promenance: int = 3) -> np.ndarray:
    """Calculates local minima (maxima) according to a prominence criterion.

    The point should be lower than any in the region around it.

    Rather than searching manually, we perform some fancy indexing to do the calculation
    across the whole array simultaneously and iterating over the promenance criterion instead of
    through the data and the promenance criterion.

    Args:
        a: The input array to calculate local minima over
        promenance: The prominence over indices required to be called a local minimum

    Returns:
        A mask where the local minima are True and other values are False.
    """
    conditions = a == a
    for i in range(1, promenance + 1):
        current_conditions = np.r_[[False] * i, a[i:] < a[:-i]] & np.r_[a[:-i] < a[i:], [False] * i]
        conditions = conditions & current_conditions

    return conditions


def local_maxima(a, promenance=3):
    return local_minima(-a, promenance)


local_maxima.__doc__ = local_minima.__doc__


def approximate_core_levels(data: DataType, window_size=None, order=5, binning=3, promenance=5):
    """Approximately locates core levels in a spectrum.

    Data is first smoothed, and then local maxima with sufficient prominence over
    other nearby points are selected as peaks.

    This can be helfpul to "seed" a curve fitting analysis for XPS.

    Args:
        data: An XPS spectrum.
        window_size: Savitzky-Golay window size
        order: Savitzky-Golay order
        binning: Used for approximate smoothing
        promenance: Required promenance over nearby peaks

    Returns:
        A set of energies with candidate peaks.
    """
    data = normalize_to_spectrum(data)

    dos = data.S.sum_other(["eV"]).sel(eV=slice(None, -20))

    if window_size is None:
        window_size = int(len(dos) / 40)  # empirical, may change
        if window_size % 2 == 0:
            window_size += 1

    smoothed = rebin(savitzky_golay(dos, window_size, order), eV=binning)

    indices = np.argwhere(local_maxima(smoothed.values, promenance=promenance))
    energies = [smoothed.coords["eV"][idx].item() for idx in indices]

    return energies



"""Very basic, generic time-resolved ARPES analysis tools."""
import numpy as np

from arpes.preparation import normalize_dim
from arpes.provenance import update_provenance
from arpes.typing import DataType
from arpes.utilities import normalize_to_spectrum

__all__ = ("find_t0", "relative_change", "normalized_relative_change")


@update_provenance("Normalized subtraction map")
def normalized_relative_change(
    data: DataType, t0=None, buffer=0.3, normalize_delay=True
) -> DataType:
    """Calculates a normalized relative Tr-ARPES change in a delay scan.

    Obtained by normalizing along the pump-probe "delay" axis and then subtracting
    the mean before t0 data and dividing by the original spectrum.

    Args:
        data: The input spectrum to be normalized. Should have a "delay" dimension.
        t0: The t0 for the input array.
        buffer: How far before t0 to select equilibrium data. Should be at least
          the temporal resolution in ps.
        normalize_delay: If true, normalizes data along the "delay" dimension.

    Returns:
        The normalized data.
    """
    spectrum = normalize_to_spectrum(data)
    if normalize_delay:
        spectrum = normalize_dim(spectrum, "delay")
    subtracted = relative_change(spectrum, t0, buffer, normalize_delay=False)
    normalized = subtracted / spectrum
    normalized.values[np.isinf(normalized.values)] = 0
    normalized.values[np.isnan(normalized.values)] = 0
    return normalized


@update_provenance("Created simple subtraction map")
def relative_change(data: DataType, t0=None, buffer=0.3, normalize_delay=True) -> DataType:
    """Like normalized_relative_change, but only subtracts the before t0 data.

    Args:
        data: The input spectrum to be normalized. Should have a "delay" dimension.
        t0: The t0 for the input array.
        buffer: How far before t0 to select equilibrium data. Should be at least
          the temporal resolution in ps.
        normalize_delay: If true, normalizes data along the "delay" dimension.

    Returns:
        The normalized data.
    """
    spectrum = normalize_to_spectrum(data)
    if normalize_delay:
        spectrum = normalize_dim(spectrum, "delay")

    delay_coords = spectrum.coords["delay"]
    delay_start = np.min(delay_coords)

    if t0 is None:
        t0 = spectrum.S.t0 or find_t0(spectrum)

    assert t0 - buffer > delay_start

    before_t0 = spectrum.sel(delay=slice(None, t0 - buffer))
    subtracted = spectrum - before_t0.mean("delay")
    return subtracted


def find_t0(data: DataType, e_bound=0.02) -> float:
    """Finds the effective t0 by fitting excited carriers.

    Args:
        data: A spectrum with "eV" and "delay" dimensions.
        e_bound: Lower bound on the energy to use for the fitting

    Returns:
        The delay value at the estimated t0.

    """
    spectrum = normalize_to_spectrum(data)

    assert "delay" in spectrum.dims
    assert "eV" in spectrum.dims

    sum_dims = set(spectrum.dims)
    sum_dims.remove("delay")
    sum_dims.remove("eV")

    summed = spectrum.sum(list(sum_dims)).sel(eV=slice(e_bound, None)).mean("eV")
    coord_max = summed.argmax().item()

    return summed.coords["delay"].values[coord_max]



"""Allows for making any function of a spectrum into a dynamic tool."""

import inspect

from PyQt5 import QtWidgets

from arpes.utilities import normalize_to_spectrum, group_by
from arpes.typing import DataType

from arpes.utilities.qt import qt_info, SimpleApp, SimpleWindow, BasicHelpDialog
from arpes.utilities.qt.ui import (
    CollectUI,
    tabs,
    horizontal,
    numeric_input,
    line_edit,
    label,
    vertical,
)

__all__ = ("make_dynamic",)

qt_info.setup_pyqtgraph()


class DynamicToolWindow(SimpleWindow):
    HELP_DIALOG_CLS = BasicHelpDialog


class DynamicTool(SimpleApp):
    WINDOW_SIZE = (
        5,
        6.5,
    )  # 5 inches by 5 inches
    WINDOW_CLS = DynamicToolWindow
    TITLE = ""  # we will use the function name for the window title

    def __init__(self, function, meta=None):
        self._function = function
        self.main_layout = QtWidgets.QGridLayout()
        self.content_layout = QtWidgets.QGridLayout()
        self.meta = meta or {}
        self.current_arguments = {}

        super().__init__()

    def layout(self):
        return self.main_layout

    def configure_image_widgets(self):
        self.generate_marginal_for(
            (), 0, 0, "xy", cursors=False, layout=self.content_layout
        )
        self.generate_marginal_for(
            (), 1, 0, "f(xy)", cursors=False, layout=self.content_layout
        )
        self.main_layout.addLayout(self.content_layout, 0, 0)

    def update_data(self):
        self.views["xy"].setImage(self.data.S.nan_to_num())
        try:
            mapped_data = self._function(self.data, **self.current_arguments)
            self.views["f(xy)"].setImage(mapped_data.S.nan_to_num())
        except:
            pass

    def add_controls(self):
        specification = self.calculate_control_specification()

        ui = {}
        with CollectUI(ui):
            controls = tabs(
                [
                    "Controls",
                    horizontal(
                        *[
                            vertical(
                                *[
                                    vertical(label(s[0]), self.build_control_for(*s))
                                    for s in pair
                                ]
                            )
                            for pair in group_by(2, specification)
                        ]
                    ),
                ],
            )

        def update_argument(arg_name, arg_type):
            def updater(value):
                self.current_arguments[arg_name] = arg_type(value)
                self.update_data()

            return updater

        for arg_name, arg_type, _ in specification:
            ui[f"{arg_name}-control"].subject.subscribe(
                update_argument(arg_name, arg_type)
            )

        controls.setFixedHeight(qt_info.inches_to_px(1.4))
        self.main_layout.addWidget(controls, 1, 0)

    def calculate_control_specification(self):
        argspec = inspect.getfullargspec(self._function)

        # we assume that the first argument is the input data
        args = argspec.args[1:]

        defaults_for_type = {
            float: 0.0,
            int: 0,
            str: "",
        }

        specs = []
        for i, arg in enumerate(args[::-1]):
            argument_type = argspec.annotations.get(arg, float)
            if i < len(argspec.defaults):
                argument_default = argspec.defaults[len(argspec.defaults) - (i + 1)]
            else:
                argument_default = defaults_for_type.get(argument_type, 0)

            self.current_arguments[arg] = argument_default
            specs.append(
                [
                    arg,
                    argument_type,
                    argument_default,
                ]
            )

        return specs

    def build_control_for(self, parameter_name, parameter_type, parameter_default):
        meta = self.meta.get(parameter_name, {})
        if parameter_type in (
            int,
            float,
        ):
            config = {}
            if "min" in meta:
                config["bottom"] = meta["min"]
            if "max" in meta:
                config["top"] = meta["max"]
            return numeric_input(
                parameter_default,
                parameter_type,
                validator_settings=config,
                id=f"{parameter_name}-control",
            )

        if parameter_type == str:
            return line_edit(parameter_default, id=f"{parameter_name}-control")

    def before_show(self):
        self.configure_image_widgets()
        self.add_controls()
        self.update_data()
        self.window.setWindowTitle(f"Interactive {self._function.__name__}")

    def set_data(self, data: DataType):
        self.data = normalize_to_spectrum(data)


def make_dynamic(fn, data):
    """Starts a tool which makes any analysis function dynamic."""
    tool = DynamicTool(fn)
    tool.set_data(data)
    tool.start()



"""Contains utility classes for Qt in PyARPES."""

import enum
from dataclasses import dataclass
from PyQt5 import QtWidgets

__all__ = ["PlotOrientation", "ReactivePlotRecord"]


class PlotOrientation(str, enum.Enum):
    """Controls the transposition on a reactive plot."""

    Horizontal = "horizontal"
    Vertical = "vertical"


@dataclass
class ReactivePlotRecord:
    """
    This contains metadata related to a reactive plot or marginal on a DataArary. This
    is used to know how to update and mount corresponding widgets on a main tool view.
    """

    dims: list[str]
    view: QtWidgets.QWidget
    orientation: PlotOrientation



"""Coordinate conversion classes for photon energy scans."""
import numpy as np
import numba

import arpes.constants
from typing import Any, Callable, Dict

from .base import CoordinateConverter, K_SPACE_BORDER, MOMENTUM_BREAKPOINTS
from .bounds_calculations import calculate_kp_kz_bounds

__all__ = ["ConvertKpKzV0", "ConvertKxKyKz", "ConvertKpKz"]


@numba.njit(parallel=True, cache=True)
def _kspace_to_hv(kp, kz, hv, energy_shift, is_constant_shift):
    """Efficiently perform the inverse coordinate transform to photon energy."""
    shift_ratio = 0 if is_constant_shift else 1

    for i in numba.prange(len(kp)):
        hv[i] = (
            arpes.constants.HV_CONVERSION * (kp[i] ** 2 + kz[i] ** 2)
            + energy_shift[i * shift_ratio]
        )


@numba.njit(parallel=True, cache=True)
def _kp_to_polar(kinetic_energy, kp, phi, inner_potential, angle_offset):
    """Efficiently performs the inverse coordinate transform phi(hv, kp)."""
    for i in numba.prange(len(kp)):
        phi[i] = (
            np.arcsin(
                kp[i]
                / (
                    arpes.constants.K_INV_ANGSTROM
                    * np.sqrt(kinetic_energy[i] + inner_potential)
                )
            )
            + angle_offset
        )


class ConvertKpKzV0(CoordinateConverter):
    """Implements inner potential broadcasted hv Fermi surfaces."""

    # TODO implement
    def __init__(self, *args, **kwargs):
        """TODO, implement this."""
        super(ConvertKpKzV0, self).__init__(*args, **kwargs)
        raise NotImplementedError


class ConvertKxKyKz(CoordinateConverter):
    """Implements 4D data volume conversion."""

    def __init__(self, *args, **kwargs):
        """TODO, implement this."""
        super(ConvertKxKyKz, self).__init__(*args, **kwargs)
        raise NotImplementedError


class ConvertKpKz(CoordinateConverter):
    """Implements single angle photon energy scans."""

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        """Cache the photon energy coordinate we calculate backwards from kz."""
        super(ConvertKpKz, self).__init__(*args, **kwargs)
        self.hv = None
        self.phi = None

    def get_coordinates(
        self, resolution: dict = None, bounds: dict = None
    ) -> Dict[str, np.ndarray]:
        """Calculates appropriate coordinate bounds."""
        if resolution is None:
            resolution = {}
        if bounds is None:
            bounds = {}

        coordinates = super(ConvertKpKz, self).get_coordinates(
            resolution=resolution, bounds=bounds
        )

        ((kp_low, kp_high), (kz_low, kz_high)) = calculate_kp_kz_bounds(self.ds)
        if "kp" in bounds:
            kp_low, kp_high = bounds["kp"]

        if "kz" in bounds:
            kz_low, kz_high = bounds["kz"]

        inferred_kp_res = (kp_high - kp_low + 2 * K_SPACE_BORDER) / len(
            self.ds.coords["phi"]
        )
        inferred_kp_res = [b for b in MOMENTUM_BREAKPOINTS if b < inferred_kp_res][-1]

        # go a bit finer here because it would otherwise be very coarse
        inferred_kz_res = (kz_high - kz_low + 2 * K_SPACE_BORDER) / len(
            self.ds.coords["hv"]
        )
        inferred_kz_res = [b for b in MOMENTUM_BREAKPOINTS if b < inferred_kz_res][-1]

        coordinates["kp"] = np.arange(
            kp_low - K_SPACE_BORDER,
            kp_high + K_SPACE_BORDER,
            resolution.get("kp", inferred_kp_res),
        )
        coordinates["kz"] = np.arange(
            kz_low - K_SPACE_BORDER,
            kz_high + K_SPACE_BORDER,
            resolution.get("kz", inferred_kz_res),
        )

        base_coords = {
            k: v for k, v in self.ds.coords.items() if k not in ["eV", "phi", "hv"]
        }

        coordinates.update(base_coords)

        return coordinates

    def kspace_to_hv(
        self,
        binding_energy: np.ndarray,
        kp: np.ndarray,
        kz: np.ndarray,
        *args: Any,
        **kwargs: Any
    ) -> np.ndarray:
        """Converts from momentum back to the raw photon energy."""
        if self.hv is None:
            inner_v = self.ds.S.inner_potential
            wf = self.ds.S.work_function

            is_constant_shift = True
            if not isinstance(binding_energy, np.ndarray):
                is_constant_shift = True
                binding_energy = np.array([binding_energy])

            self.hv = np.zeros_like(kp)
            _kspace_to_hv(
                kp, kz, self.hv, -inner_v - binding_energy + wf, is_constant_shift
            )

        return self.hv

    def kspace_to_phi(
        self,
        binding_energy: np.ndarray,
        kp: np.ndarray,
        kz: np.ndarray,
        *args: Any,
        **kwargs: Any
    ) -> np.ndarray:
        """Converts from momentum back to the hemisphere angle axis."""
        if self.phi is not None:
            return self.phi

        if self.hv is None:
            self.kspace_to_hv(binding_energy, kp, kz, *args, **kwargs)

        kinetic_energy = binding_energy + self.hv - self.ds.S.work_function

        self.phi = np.zeros_like(self.hv)

        _kp_to_polar(
            kinetic_energy,
            kp,
            self.phi,
            self.ds.S.inner_potential,
            self.ds.S.phi_offset,
        )

        try:
            self.phi = self.calibration.correct_detector_angle(
                eV=binding_energy, phi=self.phi
            )
        except:
            pass

        return self.phi

    def conversion_for(self, dim: str) -> Callable:
        """Looks up the appropriate momentum-to-angle conversion routine by dimension name."""

        def with_identity(*args, **kwargs):
            return self.identity_transform(dim, *args, **kwargs)

        return {
            "eV": self.kspace_to_BE,
            "hv": self.kspace_to_hv,
            "phi": self.kspace_to_phi,
        }.get(dim, with_identity)



"""Plotting routines related to 2D ARPES cuts and dispersions."""
import warnings
from collections import defaultdict

import matplotlib.pyplot as plt
import numpy as np

import xarray as xr
from arpes.io import load_data
from arpes.preparation import normalize_dim
from arpes.provenance import save_plot_provenance
from arpes.utilities import bz
from arpes.utilities.conversion import remap_coords_to

from .utils import path_for_plot, label_for_dim, label_for_symmetry_point, label_for_colorbar

__all__ = [
    "plot_dispersion",
    "labeled_fermi_surface",
    "cut_dispersion_plot",
    "fancy_dispersion",
    "reference_scan_fermi_surface",
    "hv_reference_scan",
    "scan_var_reference_plot",
]


@save_plot_provenance
def plot_dispersion(spectrum: xr.DataArray, bands, out=None):
    """Plots an ARPES cut with bands over it."""
    ax = spectrum.plot()

    for band in bands:
        plt.scatter(band.center.values, band.coords[band.dims[0]].values)

    if out is not None:
        filename = path_for_plot(out)
        plt.savefig(filename)
        return filename
    else:
        return ax


@save_plot_provenance
def cut_dispersion_plot(
    data: xr.DataArray,
    e_floor=None,
    title=None,
    ax=None,
    include_symmetry_points=True,
    out=None,
    quality="high",
    **kwargs
):
    """Makes a 3D cut dispersion plot.

    At the moment this only supports rectangular BZs.

    Args:
        data: The 3D data to plot
        e_floor: The energy of the bottom cut "floor"
        title: A title for the plot
        ax: The axes to plot to
        include_symmetry_points: Whether to include annotated symmetry points
        out: Where to save the file, optionall
        quality: Controls output figure DPI
        kwargs: Fed to the image plotting calls
    """
    # to get nice labeled edges you could use shapely
    sampling = {
        "paper": 400,
        "high": 100,
        "low": 40,
    }.get(quality, 100)

    assert "eV" in data.dims
    assert e_floor is not None

    new_dim_order = list(data.dims)
    new_dim_order.remove("eV")
    new_dim_order = new_dim_order + ["eV"]
    data = data.transpose(*new_dim_order)

    # prep data to be used for rest of cuts
    lower_part = data.sel(eV=slice(None, 0))
    floor = lower_part.S.fat_sel(eV=e_floor)

    bz_mask = bz.reduced_bz_mask(lower_part, scale_zone=True)
    left_mask = bz.reduced_bz_E_mask(lower_part, "X", e_floor, scale_zone=True)
    right_mask = bz.reduced_bz_E_mask(lower_part, "Y", e_floor, scale_zone=True)
    mask_for = lambda x: left_mask if x.shape == left_mask.shape else right_mask

    x_dim, y_dim, z_dim = tuple(new_dim_order)
    x_coords, y_coords, _ = (
        data.coords[x_dim],
        data.coords[y_dim],
        data.coords[z_dim],
    )  # x_coords, y_coords, z_coords

    if ax is None:
        fig = plt.figure(figsize=(7, 7))
        ax = fig.add_subplot(1, 1, 1, projection="3d")

    if title is None:
        title = "{} Cut Through Symmetry Points".format(data.S.label.replace("_", " "))

    ax.set_title(title)
    colormap = plt.get_cmap("Blues")

    # color fermi surface
    fermi_surface = data.S.fermi_surface
    Xs, Ys = np.meshgrid(x_coords, y_coords)

    Zs = np.zeros(fermi_surface.data.shape)
    Zs[bz_mask] = np.nan
    scale_colors = max(np.max(fermi_surface.data), np.max(floor.data))
    ax.plot_surface(
        Xs,
        Ys,
        Zs.T,
        facecolors=colormap(fermi_surface.data.T / scale_colors),
        shade=False,
        vmin=-1,
        vmax=1,
        antialiased=True,
        rcount=sampling,
        ccount=sampling,
    )

    # color right edge
    right_sel = dict()
    edge_val = np.max(lower_part.coords[x_dim].data)
    right_sel[x_dim] = edge_val
    right_edge = lower_part.S.fat_sel(**right_sel)

    Ys, Zs = np.meshgrid(lower_part.coords[y_dim], lower_part.coords[z_dim])
    Xs = np.ones(right_edge.shape) * edge_val
    Xs[mask_for(Xs)] = np.nan
    ax.plot_surface(
        Xs.T,
        Ys,
        Zs,
        facecolors=colormap(right_edge.data.T / scale_colors),
        vmin=-1,
        vmax=1,
        shade=False,
        antialiased=True,
        rcount=sampling,
        ccount=sampling,
    )

    # color left edge
    left_sel = dict()
    edge_val = np.min(lower_part.coords[y_dim].data)
    left_sel[y_dim] = edge_val
    left_edge = lower_part.S.fat_sel(**left_sel)
    # max_left_edge = np.max(left_edge.data)

    Xs, Zs = np.meshgrid(lower_part.coords[x_dim], lower_part.coords[z_dim])
    Ys = np.ones(left_edge.shape) * edge_val
    Ys[mask_for(Ys)] = np.nan
    ax.plot_surface(
        Xs,
        Ys.T,
        Zs,
        facecolors=colormap(left_edge.data.T / scale_colors),
        vmin=-1,
        vmax=1,
        antialiased=True,
        shade=False,
        rcount=sampling,
        ccount=sampling,
    )

    # selection region
    # floor
    Xs, Ys = np.meshgrid(floor.coords[x_dim], floor.coords[y_dim])
    Zs = np.ones(floor.data.shape) * e_floor
    Zs[np.logical_not(bz_mask)] = np.nan
    ax.plot_surface(
        Xs,
        Ys,
        Zs.T,
        facecolors=colormap(floor.data.T / scale_colors),
        vmin=-1,
        vmax=1,
        shade=False,
        antialiased=True,
        rcount=sampling,
        ccount=sampling,
    )

    # determine the axis along X, Y
    axis_X = bz.axis_along(data, "X")
    axis_Y = bz.axis_along(data, "Y")

    # left and right inset faces
    inset_face = lower_part.S.along(["G", "X"], axis_name=axis_X, extend_to_edge=True).sel(
        eV=slice(e_floor, None)
    )
    Xs, Zs = np.meshgrid(inset_face.coords[axis_X], inset_face.coords[z_dim])
    Ys = np.ones(inset_face.data.shape)
    if y_dim == axis_X:
        Ys *= data.S.phi_offset
        Xs += data.S.map_angle_offset
        Xs, Ys = Ys, Xs
    else:
        Ys *= data.S.map_angle_offset
        Xs += data.S.phi_offset
    ax.plot_surface(
        Xs,
        Ys,
        Zs,
        facecolors=colormap(inset_face.data / scale_colors),
        shade=False,
        antialiased=True,
        zorder=1,
        rcount=sampling,
        ccount=sampling,
    )

    inset_face = lower_part.S.along(["G", "Y"], axis_name=axis_Y, extend_to_edge=True).sel(
        eV=slice(e_floor, None)
    )
    Ys, Zs = np.meshgrid(inset_face.coords[axis_Y], inset_face.coords[z_dim])
    Xs = np.ones(inset_face.data.shape)
    if x_dim == axis_Y:
        Xs *= data.S.map_angle_offset
        Ys += data.S.phi_offset
        Xs, Ys = Ys, Xs
    else:
        Xs *= data.S.phi_offset
        Ys += data.S.map_angle_offset

    ax.plot_surface(
        Xs,
        Ys,
        Zs,
        facecolors=colormap(inset_face.data / scale_colors),
        shade=False,
        antialiased=True,
        zorder=1,
        rcount=sampling,
        ccount=sampling,
    )

    ax.set_xlabel(label_for_dim(data, x_dim))
    ax.set_ylabel(label_for_dim(data, y_dim))
    ax.set_zlabel(label_for_dim(data, z_dim))

    zlim = ax.get_zlim3d()
    if include_symmetry_points:
        for point_name, point_location in data.S.iter_symmetry_points:
            coords = [point_location.get(d, 0.02) for d in new_dim_order]
            ax.scatter(*zip(coords), marker=".", color="red", zorder=1000)
            coords[new_dim_order.index("eV")] += 0.1
            ax.text(
                *coords, label_for_symmetry_point(point_name), color="red", ha="center", va="top"
            )

    ax.set_zlim3d(*zlim)
    if out is not None:
        plt.savefig(path_for_plot(out), dpi=400)
        return path_for_plot(out)

    plt.show()


@save_plot_provenance
def hv_reference_scan(data, out=None, e_cut=-0.05, bkg_subtraction=0.8, **kwargs):
    """A reference plot for photon energy scans. Used internally by other code."""
    fs = data.S.fat_sel(eV=e_cut)
    fs = normalize_dim(fs, "hv", keep_id=True)
    fs.data -= bkg_subtraction * np.mean(fs.data)
    fs.data[fs.data < 0] = 0

    _, ax = labeled_fermi_surface(fs, hold=True, **kwargs)

    all_scans = data.attrs["df"]
    all_scans = all_scans[all_scans.id != data.attrs["id"]]
    all_scans = all_scans[
        (all_scans.spectrum_type != "xps_spectrum") | (all_scans.spectrum_type == "hv_map")
    ]

    scans_by_hv = defaultdict(list)
    for _, row in all_scans.iterrows():
        scan = load_data(row.id)

        scans_by_hv[round(scan.S.hv)].append(scan.S.label.replace("_", " "))

    dim_order = ax.dim_order
    handles = []
    handle_labels = []

    prop_cycle = plt.rcParams["axes.prop_cycle"]
    colors_cycle = prop_cycle.by_key()["color"]

    for line_color, (hv, labels) in zip(colors_cycle, scans_by_hv.items()):
        full_label = "\n".join(labels)

        # determine direction
        if dim_order[0] == "hv":
            # photon energy is along the x axis, we want an axvline
            handles.append(ax.axvline(hv, label=full_label, color=line_color))
        else:
            # photon energy is along the y axis, we want an axhline
            handles.append(ax.axhline(hv, label=full_label, color=line_color))

        handle_labels.append(full_label)

    plt.legend(handles, handle_labels)

    if out is not None:
        plt.savefig(path_for_plot(out), dpi=400)
        return path_for_plot(out)

    plt.show()


@save_plot_provenance
def reference_scan_fermi_surface(data, out=None, **kwargs):
    """A reference plot for Fermi surfaces. Used internally by other code."""
    fs = data.S.fermi_surface
    _, ax = labeled_fermi_surface(fs, hold=True, **kwargs)

    referenced_scans = data.S.referenced_scans
    handles = []
    for index, row in referenced_scans.iterrows():
        scan = load_data(row.id)
        remapped_coords = remap_coords_to(scan, data)
        dim_order = ax.dim_order
        ls = ax.plot(
            remapped_coords[dim_order[0]],
            remapped_coords[dim_order[1]],
            label=index.replace("_", " "),
        )
        handles.append(ls[0])

    plt.legend(handles=handles)

    if out is not None:
        plt.savefig(path_for_plot(out), dpi=400)
        return path_for_plot(out)

    plt.show()


@save_plot_provenance
def labeled_fermi_surface(
    data,
    title=None,
    ax=None,
    hold=False,
    include_symmetry_points=True,
    include_bz=True,
    out=None,
    fermi_energy=0,
    **kwargs
):
    """Plots a Fermi surface with high symmetry points annotated onto it."""
    fig = None
    if ax is None:
        fig, ax = plt.subplots(figsize=(7, 7))

    if title is None:
        title = "{} Fermi Surface".format(data.S.label.replace("_", " "))

    if "eV" in data.dims:
        data = data.S.generic_fermi_surface(fermi_energy)

    mesh = data.plot(ax=ax)
    mesh.colorbar.set_label(label_for_colorbar(data))

    if data.S.is_differentiated:
        mesh.set_cmap("Blues")

    dim_order = [ax.get_xlabel(), ax.get_ylabel()]

    setattr(ax, "dim_order", dim_order)
    ax.set_xlabel(label_for_dim(data, ax.get_xlabel()))
    ax.set_ylabel(label_for_dim(data, ax.get_ylabel()))
    ax.set_title(title)

    marker_color = "red" if data.S.is_differentiated else "red"

    if include_bz:
        symmetry = bz.bz_symmetry(data.S.iter_own_symmetry_points)

        # TODO Implement this
        warnings.warn("BZ region display not implemented.")

    if include_symmetry_points:
        for point_name, point_location in data.S.iter_symmetry_points:
            warnings.warn("Symmetry point locations are not k-converted")
            coords = [point_location[d] for d in dim_order]
            ax.plot(*coords, marker=".", color=marker_color)
            ax.annotate(
                label_for_symmetry_point(point_name),
                coords,
                color=marker_color,
                xycoords="data",
                textcoords="offset points",
                xytext=(0, -10),
                va="top",
                ha="center",
                fontsize=14,
            )

    if out is not None:
        plt.savefig(path_for_plot(out), dpi=400)
        return path_for_plot(out)

    if not hold:
        plt.show()
    else:
        return fig, ax


@save_plot_provenance
def fancy_dispersion(
    data, title=None, ax=None, out=None, include_symmetry_points=True, norm=None, **kwargs
):
    """Generates a 2D ARPES cut with some fancy annotations for throwing plots together.

    Useful for brief slides/quick presentations.
    """
    if ax is None:
        _, ax = plt.subplots(figsize=(8, 5))

    if title is None:
        title = data.S.label.replace("_", " ")

    mesh = data.plot(norm=norm, ax=ax, **kwargs)
    mesh.colorbar.set_label(label_for_colorbar(data))

    if data.S.is_differentiated:
        mesh.set_cmap("Blues")

    original_x_label = ax.get_xlabel()
    ax.set_xlabel(label_for_dim(data, ax.get_xlabel()))
    ax.set_ylabel(label_for_dim(data, ax.get_ylabel()))
    ax.set_title(title, fontsize=14)

    # This can probably be pulled out into a a helper
    marker_color = "red" if data.S.is_differentiated else "red"
    if include_symmetry_points:
        for point_name, point_locations in data.attrs.get("symmetry_points", {}).items():
            if not isinstance(point_locations, list):
                point_locations = [point_locations]
            for single_location in point_locations:
                coords = (
                    single_location[original_x_label],
                    ax.get_ylim()[1],
                )
                ax.plot(*coords, marker=11, color=marker_color)
                ax.annotate(
                    label_for_symmetry_point(point_name),
                    coords,
                    color=marker_color,
                    xycoords="data",
                    textcoords="offset points",
                    xytext=(0, -10),
                    va="top",
                    ha="center",
                )

    ax.axhline(0, color="red", alpha=0.8, linestyle="--", linewidth=1)

    if out is not None:
        plt.savefig(path_for_plot(out), dpi=400)
        return path_for_plot(out)

    plt.show()
    return ax


@save_plot_provenance
def scan_var_reference_plot(data, title=None, ax=None, norm=None, out=None, **kwargs):
    """Makes a straightforward plot of a DataArray with resonable axes. Used internally by other scripts."""
    if ax is None:
        _, ax = plt.subplots(figsize=(8, 5))

    if title is None:
        title = data.S.label.replace("_", " ")

    plot = data.plot(norm=norm, ax=ax)
    plot.colorbar.set_label(label_for_colorbar(data))

    ax.set_xlabel(label_for_dim(data, ax.get_xlabel()))
    ax.set_ylabel(label_for_dim(data, ax.get_ylabel()))

    ax.set_title(title, font_size=14)

    if out is not None:
        plt.savefig(path_for_plot(out), dpi=400)
        return path_for_plot(out)

    plt.show()



"""Implements 2D and 3D angle scan momentum conversion for Fermi surfaces.

Broadly, this covers cases where we are not performing photon energy scans.
"""

import numpy as np

import numba
import math

import arpes.constants
import xarray as xr
from typing import Any, Callable

from .base import CoordinateConverter, K_SPACE_BORDER, MOMENTUM_BREAKPOINTS
from .bounds_calculations import calculate_kp_bounds, calculate_kx_ky_bounds

__all__ = ["ConvertKp", "ConvertKxKy"]


@numba.njit(parallel=True)
def _exact_arcsin(k_par, k_perp, k_tot, phi, offset, par_tot, negate):
    """A efficient arcsin with total momentum scaling."""
    mul_idx = 1 if par_tot else 0
    for i in numba.prange(len(k_par)):
        result = np.arcsin(k_par[i] / np.sqrt(k_tot[i * mul_idx] ** 2 - k_perp[i] ** 2))
        if negate:
            result = -result

        phi[i] = result + offset


@numba.njit(parallel=True)
def _small_angle_arcsin(k_par, k_tot, phi, offset, par_tot, negate):
    """A efficient small angle arcsin with total momentum scaling.

    np.arcsin(k_par / k_tot, phi)
    phi += offset
    mul_idx = 0
    """
    mul_idx = 1 if par_tot else 0
    for i in numba.prange(len(k_par)):
        result = np.arcsin(k_par[i] / k_tot[i * mul_idx])

        if negate:
            result = -result

        phi[i] = result + offset


@numba.njit(parallel=True)
def _rotate_kx_ky(kx, ky, kxout, kyout, chi):
    cos_chi = np.cos(chi)
    sin_chi = np.sin(chi)
    for i in numba.prange(len(kx)):
        kxout[i] = kx[i] * cos_chi - ky[i] * sin_chi
        kyout[i] = ky[i] * cos_chi + kx[i] * sin_chi


@numba.njit(parallel=True)
def _compute_ktot(hv, work_function, binding_energy, k_tot):
    for i in numba.prange(len(binding_energy)):
        k_tot[i] = arpes.constants.K_INV_ANGSTROM * math.sqrt(
            hv - work_function + binding_energy[i]
        )


def _safe_compute_k_tot(hv, work_function, binding_energy):
    arr_binding_energy = binding_energy
    if not isinstance(binding_energy, np.ndarray):
        arr_binding_energy = np.array([binding_energy])

    k_tot = np.zeros_like(arr_binding_energy)
    _compute_ktot(hv, work_function, arr_binding_energy, k_tot)

    return k_tot


class ConvertKp(CoordinateConverter):
    """A momentum converter for single ARPES (kp) cuts."""

    k_tot: np.ndarray = None
    phi: np.ndarray = None

    def __init__(self, ds: xr.Dataset, *args: Any, **kwargs: Any) -> None:
        """Initialize the kp momentum converter and cached coordinate values."""
        super().__init__(ds, *args, **kwargs)

        if self.is_slit_vertical:
            self.parallel_angle = "beta"
            self.perpendicular_angles = ["psi", "theta"]
        else:
            self.parallel_angle = "theta"
            self.perpendicular_angles = ["psi", "beta"]

    @property
    def phi_offset(self) -> float:
        """The offset for the phi angle."""
        return self.ds.S.phi_offset + self.ds.S.lookup_offset_coord(self.parallel_angle)

    @property
    def perpendicular_offset(self) -> float:
        """The offset for the perpendicular angle."""
        return sum(
            self.ds.S.lookup_offset_coord(angle) for angle in self.perpendicular_angles
        )

    def get_coordinates(
        self, resolution: dict = None, bounds: dict = None
    ) -> dict[str, np.ndarray]:
        """Calculates appropriate coordinate bounds."""
        resolution = {} if resolution is None else resolution
        bounds = {} if bounds is None else bounds

        coordinates = super().get_coordinates(resolution, bounds=bounds)
        if "kp" in bounds:
            kp_low, kp_high = bounds["kp"]
        else:
            kp_low, kp_high = calculate_kp_bounds(
                self.ds, self.phi_offset, self.perpendicular_offset
            )

        inferred_kp_res = (kp_high - kp_low + 2 * K_SPACE_BORDER) / len(
            self.ds.coords["phi"]
        )

        try:
            inferred_kp_res = [b for b in MOMENTUM_BREAKPOINTS if b < inferred_kp_res][
                -2 if (len(self.ds.coords["phi"]) < 80) else -1
            ]
        except IndexError:
            inferred_kp_res = MOMENTUM_BREAKPOINTS[-2]

        coordinates["kp"] = np.arange(
            kp_low - K_SPACE_BORDER,
            kp_high + K_SPACE_BORDER,
            resolution.get("kp", inferred_kp_res),
        )

        base_coords = {
            k: v
            for k, v in self.ds.coords.items()
            if k not in ["eV", "phi", "beta", "theta"]
        }

        coordinates.update(base_coords)
        return coordinates

    def compute_k_tot(self, binding_energy: np.ndarray) -> None:
        """Compute the total momentum (inclusive of kz) at different binding energies."""
        self.k_tot = _safe_compute_k_tot(
            self.ds.S.hv, self.ds.S.work_function, binding_energy
        )

    def kspace_to_phi(
        self, binding_energy: np.ndarray, kp: np.ndarray, *args: Any, **kwargs: Any
    ) -> np.ndarray:
        """Converts from momentum back to the analyzer angular axis."""
        if self.phi is not None:
            return self.phi

        if self.k_tot is None:
            self.compute_k_tot(binding_energy)

        self.phi = np.zeros_like(kp)
        par_tot = isinstance(self.k_tot, np.ndarray) and len(self.k_tot) != 1
        assert len(self.k_tot) == len(kp) or len(self.k_tot) == 1

        _small_angle_arcsin(
            kp / np.cos(self.perpendicular_offset),
            self.k_tot,
            self.phi,
            self.phi_offset,
            par_tot,
            False,
        )

        if self.calibration is not None:
            self.phi = self.calibration.correct_detector_angle(
                eV=binding_energy, phi=self.phi
            )
        return self.phi

    def conversion_for(self, dim: str) -> Callable:
        """Looks up the appropriate momentum-to-angle conversion routine by dimension name."""

        def with_identity(*args, **kwargs):
            return self.identity_transform(dim, *args, **kwargs)

        return {"eV": self.kspace_to_BE, "phi": self.kspace_to_phi}.get(
            dim, with_identity
        )


class ConvertKxKy(CoordinateConverter):
    """Implements volumetric momentum conversion for kx-ky scans.

    Please note that currently we assume that psi = 0 when you are not using an
    electrostatic deflector.
    """

    k_tot: np.ndarray = None
    phi: np.ndarray = None
    perp_angle: np.ndarray = None
    rkx: np.ndarray = None
    rky: np.ndarray = None

    def __init__(self, ds: xr.Dataset, *args: list[str], **kwargs: Any) -> None:
        """Initialize the kx-ky momentum converter and cached coordinate values."""
        super().__init__(ds, *args, **kwargs)
        # accept either vertical or horizontal, fail otherwise
        if not any(
            np.abs(ds.alpha - alpha_option) < (np.pi / 180)
            for alpha_option in [0, np.pi / 2]
        ):
            raise ValueError(
                "You must convert either vertical or horizontal slit data with this "
                "converter."
            )

        self.direct_angles = (
            "phi",
            [d for d in ["psi", "beta", "theta"] if d in ds.indexes][0],
        )
        assert len(self.direct_angles) == 2, "couldn't find an appropriate scan angle"

        if self.direct_angles[1] != "psi":
            # psi allows for either orientation
            assert (
                self.direct_angles[1] == "beta"
            ) != self.is_slit_vertical, "slit orientation doesn't match scan angle"

        if self.is_slit_vertical:
            self.parallel_angles = (
                "beta",
                "psi" if "theta" in self.direct_angles else "theta",
            )
        else:
            self.parallel_angles = (
                "theta",
                "psi" if "beta" in self.direct_angles else "beta",
            )

    @property
    def phi_offset(self) -> float:
        """The offset for the phi angle."""
        return self.ds.S.phi_offset + self.ds.S.lookup_offset_coord(
            self.parallel_angles[0]
        )

    @property
    def perp_offset(self) -> float:
        """The offset for the perpendicular angle."""
        scan_angle = self.direct_angles[1]
        return self.ds.S.lookup_offset(scan_angle) + self.ds.S.lookup_offset_coord(
            self.parallel_angles[1]
        )

    def get_coordinates(
        self, resolution: dict = None, bounds: dict = None
    ) -> dict[str, np.ndarray]:
        """Calculates appropriate coordinate bounds."""
        resolution = {} if resolution is None else resolution
        bounds = {} if bounds is None else bounds

        coordinates = super().get_coordinates(resolution, bounds=bounds)

        kx_angle, ky_angle = self.direct_angles
        ((kx_low, kx_high), (ky_low, ky_high)) = calculate_kx_ky_bounds(
            self.ds, self.phi_offset, self.perp_offset, ky_angle
        )

        if "kx" in bounds:
            kx_low, kx_high = bounds["kx"]
        if "ky" in bounds:
            ky_low, ky_high = bounds["ky"]

        len_ky_angle = len(self.ds.coords[ky_angle])
        len_kx_angle = len(self.ds.coords[kx_angle])

        inferred_kx_res = (kx_high - kx_low + 2 * K_SPACE_BORDER) / len_kx_angle
        inferred_ky_res = (ky_high - ky_low + 2 * K_SPACE_BORDER) / len_ky_angle
        # upsample a bit if there aren't that many points along a certain axis
        try:
            inferred_kx_res = [b for b in MOMENTUM_BREAKPOINTS if b < inferred_kx_res][
                -2 if (len_kx_angle < 80) else -1
            ]
        except IndexError:
            inferred_kx_res = MOMENTUM_BREAKPOINTS[-2]
        try:
            inferred_ky_res = [b for b in MOMENTUM_BREAKPOINTS if b < inferred_ky_res][
                -2 if (len_ky_angle < 80) else -1
            ]
        except IndexError:
            inferred_ky_res = MOMENTUM_BREAKPOINTS[-2]

        coordinates["kx"] = np.arange(
            kx_low - K_SPACE_BORDER,
            kx_high + K_SPACE_BORDER,
            resolution.get("kx", inferred_kx_res),
        )
        coordinates["ky"] = np.arange(
            ky_low - K_SPACE_BORDER,
            ky_high + K_SPACE_BORDER,
            resolution.get("ky", inferred_ky_res),
        )

        # TODO J: figure out a better way to handle this
        # J: turns out this is required even while keeping the original dataset
        base_coords = {
            k: v
            for k, v in self.ds.coords.items()
            if k not in ["eV", "phi", "psi", "theta", "beta", "alpha", "chi"]
        }
        coordinates.update(base_coords)

        return coordinates

    def compute_k_tot(self, binding_energy: np.ndarray) -> None:
        """Compute the total momentum (inclusive of kz) at different binding energies."""
        self.k_tot = _safe_compute_k_tot(
            self.ds.S.hv, self.ds.S.work_function, binding_energy
        )

    def conversion_for(self, dim: str) -> Callable:
        """Looks up the appropriate momentum-to-angle conversion routine by dimension name."""

        def with_identity(*args, **kwargs):
            return self.identity_transform(dim, *args, **kwargs)

        return {
            "eV": self.kspace_to_BE,
            "phi": self.kspace_to_phi,
            "theta": self.kspace_to_perp_angle,
            "psi": self.kspace_to_perp_angle,
            "beta": self.kspace_to_perp_angle,
        }.get(dim, with_identity)

    @property
    def needs_rotation(self) -> bool:
        """Whether we need to rotate the momentum coordinates when converting to angle."""
        # force rotation when greater than 0.5 deg
        return np.abs(self.ds.S.lookup_offset_coord("chi")) > (0.5 * np.pi / 180)

    def rkx_rky(self, kx, ky):
        """Returns the rotated kx and ky values when we are rotating by nonzero chi."""
        if self.rkx is not None:
            return self.rkx, self.rky

        chi = self.ds.S.lookup_offset_coord("chi")

        self.rkx = np.zeros_like(kx)
        self.rky = np.zeros_like(ky)
        _rotate_kx_ky(kx, ky, self.rkx, self.rky, chi)

        return self.rkx, self.rky

    def kspace_to_phi(
        self,
        binding_energy: np.ndarray,
        kx: np.ndarray,
        ky: np.ndarray,
        *args: Any,
        **kwargs: Any,
    ) -> np.ndarray:
        """Converts from momentum back to the analyzer angular axis."""
        if self.phi is not None:
            return self.phi

        if self.k_tot is None:
            self.compute_k_tot(binding_energy)

        if self.needs_rotation:
            kx, ky = self.rkx_rky(kx, ky)

        # This can be condensed but it is actually better not to condense it:
        # In this format, we can very easily compare to the raw coordinate conversion functions that
        # come from Mathematica in order to adjust signs, etc.
        self.phi = np.zeros_like(ky)

        par_tot = isinstance(self.k_tot, np.ndarray) and len(self.k_tot) != 1
        assert len(self.k_tot) == len(self.phi) or len(self.k_tot) == 1

        _exact_arcsin(kx, ky, self.k_tot, self.phi, self.phi_offset, par_tot, False)

        if self.calibration is not None:
            self.phi = self.calibration.correct_detector_angle(
                eV=binding_energy, phi=self.phi
            )

        return self.phi

    def kspace_to_perp_angle(
        self,
        binding_energy: np.ndarray,
        kx: np.ndarray,
        ky: np.ndarray,
        *args: Any,
        **kwargs: Any,
    ) -> np.ndarray:
        """Convert momentum to the scan angle perpendicular to the analyzer slit."""
        if self.perp_angle is not None:
            return self.perp_angle

        if self.k_tot is None:
            self.compute_k_tot(binding_energy)

        if self.needs_rotation:
            kx, ky = self.rkx_rky(kx, ky)

        self.perp_angle = np.zeros_like(kx)

        par_tot = isinstance(self.k_tot, np.ndarray) and len(self.k_tot) != 1
        assert len(self.k_tot) == len(self.perp_angle) or len(self.k_tot) == 1

        _exact_arcsin(
            ky, kx, self.k_tot, self.perp_angle, self.perp_offset, par_tot, False
        )

        return self.perp_angle



"""Derivative, curvature, and minimum gradient analysis."""

import functools
import warnings

import numpy as np

import xarray as xr
from arpes.provenance import provenance, update_provenance
from arpes.typing import DataType
from arpes.utilities import normalize_to_spectrum, lift_spectrum


__all__ = (
    "curvature",
    "dn_along_axis",
    "d2_along_axis",
    "d1_along_axis",
    "minimum_gradient",
    "vector_diff",
)


def vector_diff(arr: np.ndarray, delta, n=1):
    """Computes finite differences along the vector delta, given as a tuple.

    Using delta = (0, 1) is equivalent to np.diff(..., axis=1), while
    using delta = (1, 0) is equivalent to np.diff(..., axis=0).

    Args:
        arr: The input array
        delta: iterable containing vector to take difference along

    Returns:
        The finite differences along the translation vector provided.
    """
    if n == 0:
        return arr
    if n < 0:
        raise ValueError("Order must be non-negative but got " + repr(n))

    nd = arr.ndim
    slice1 = [slice(None)] * nd
    slice2 = [slice(None)] * nd

    for dim, delta_val in enumerate(delta):
        if delta_val != 0:
            if delta_val < 0:
                slice2[dim] = slice(-delta_val, None)
                slice1[dim] = slice(None, delta_val)
            else:
                slice1[dim] = slice(delta_val, None)
                slice2[dim] = slice(None, -delta_val)

    slice1, slice2 = tuple(slice1), tuple(slice2)

    if n > 1:
        return vector_diff(arr[slice1] - arr[slice2], delta, n - 1)

    return arr[slice1] - arr[slice2]


@update_provenance("Minimum Gradient")
def minimum_gradient(data: DataType, delta=1):
    """Implements the minimum gradient approach to defining the band in a diffuse spectrum."""
    arr = normalize_to_spectrum(data)
    new = arr / gradient_modulus(arr, delta=delta)
    new.values[np.isnan(new.values)] = 0
    return new


@update_provenance("Gradient Modulus")
def gradient_modulus(data: DataType, delta=1):
    spectrum = normalize_to_spectrum(data)
    values = spectrum.values
    gradient_vector = np.zeros(shape=(8,) + values.shape)

    gradient_vector[0, :-delta, :] = vector_diff(values, (delta, 0))
    gradient_vector[1, :, :-delta] = vector_diff(values, (0, delta))
    gradient_vector[2, delta:, :] = vector_diff(values, (-delta, 0))
    gradient_vector[3, :, delta:] = vector_diff(values, (0, -delta))
    gradient_vector[4, :-delta, :-delta] = vector_diff(values, (delta, delta))
    gradient_vector[5, :-delta, delta:] = vector_diff(values, (delta, -delta))
    gradient_vector[6, delta:, :-delta] = vector_diff(values, (-delta, delta))
    gradient_vector[7, delta:, delta:] = vector_diff(values, (-delta, -delta))

    data_copy = spectrum.copy(deep=True)
    data_copy.values = np.linalg.norm(gradient_vector, axis=0)
    return data_copy


@lift_spectrum
def curvature(spectrum: xr.DataArray, directions=None, alpha=1, beta=None):
    r"""Provides "curvature" analysis for band locations.

    Defined via

    .. math::

        C(x,y) = \frac{([C_0 + (df/dx)^2]\frac{d^2f}{dy^2} - 2 \frac{df}{dx}\frac{df}{dy} \frac{d^2f}{dxdy} + [C_0 + (\frac{df}{dy})^2]\frac{d^2f}{dx^2})}{
            (C_0 (\frac{df}{dx})^2 + (\frac{df}{dy})^2)^{3/2}}


    of in the case of inequivalent dimensions x and y

    .. math::

        C(x,y) = \frac{[1 + C_x(\frac{df}{dx})^2]C_y
        \frac{d^2f}{dy^2} - 2 C_x  C_y  \frac{df}{dx}\frac{df}{dy}\frac{d^2f}{dxdy} +
        [1 + C_y (\frac{df}{dy})^2] C_x \frac{d^2f}{dx^2}}{
        (1 + C_x (\frac{df}{dx})^2 + C_y (\frac{df}{dy})^2)^{3/2}}

    where

    .. math::

        C_x = C_y (\frac{dx}{dy})^2

    The value of C_y can reasonably be taken to have the value

    .. math::

        (\frac{df}{dx})_\text{max}^2 + \left|\frac{df}{dy}\right|_\text{max}^2
        C_y = (\frac{dy}{dx}) (\left|\frac{df}{dx}\right|_\text{max}^2 + \left|\frac{df}{dy}\right|_\text{max}^2) \alpha

    for some dimensionless parameter :math:`\alpha`.

    Args:
        arr
        alpha: regulation parameter, chosen semi-universally, but with
            no particular justification

    Returns:
        The curvature of the intensity of the original data.
    """
    if beta is not None:
        alpha = np.power(10.0, beta)

    if directions is None:
        directions = spectrum.dims[:2]

    axis_indices = tuple(spectrum.dims.index(d) for d in directions)
    dx, dy = tuple(
        float(spectrum.coords[d][1] - spectrum.coords[d][0]) for d in directions
    )
    dfx, dfy = np.gradient(spectrum.values, dx, dy, axis=axis_indices)
    np.nan_to_num(dfx, copy=False)
    np.nan_to_num(dfy, copy=False)

    mdfdx, mdfdy = np.max(np.abs(dfx)), np.max(np.abs(dfy))

    cy = (dy / dx) * (mdfdx**2 + mdfdy**2) * alpha
    cx = (dx / dy) * (mdfdx**2 + mdfdy**2) * alpha

    dfx_2, dfy_2 = np.power(dfx, 2), np.power(dfy, 2)
    d2fy = np.gradient(dfy, dy, axis=axis_indices[1])
    d2fx = np.gradient(dfx, dx, axis=axis_indices[0])
    d2fxy = np.gradient(dfx, dy, axis=axis_indices[1])

    denom = np.power((1 + cx * dfx_2 + cy * dfy_2), 1.5)
    numerator = (
        (1 + cx * dfx_2) * cy * d2fy
        - 2 * cx * cy * dfx * dfy * d2fxy
        + (1 + cy * dfy_2) * cx * d2fx
    )

    curv = xr.DataArray(numerator / denom, spectrum.coords, spectrum.dims)

    return curv


def dn_along_axis(
    arr: xr.DataArray, axis=None, smooth_fn=None, order=2
) -> xr.DataArray:
    """Like curvature, performs a second derivative.

    You can pass a function to use for smoothing through
    the parameter smooth_fn, otherwise no smoothing will be performed.

    You can specify the axis to take the derivative along with the axis param, which expects a string.
    If no axis is provided the axis will be chosen from among the available ones according to the preference
    for axes here, the first available being taken:

    ['eV', 'kp', 'kx', 'kz', 'ky', 'phi', 'beta', 'theta]

    Args:
        arr
        axis
        smooth_fn
        order: Specifies how many derivatives to take

    Returns:
        The nth derivative data.
    """
    axis_order = ["eV", "kp", "kx", "kz", "ky", "phi", "beta", "theta"]
    if axis is None:
        axes = [a for a in axis_order if a in arr.dims]
        if axes:
            axis = axes[0]
        else:
            # have to do something
            axis = arr.dims[0]
            warnings.warn(
                "Choosing axis: {} for the second derivative, no preferred axis found.".format(
                    axis
                )
            )

    if smooth_fn is None:
        smooth_fn = lambda x: x

    d_axis = float(arr.coords[axis][1] - arr.coords[axis][0])
    axis_idx = arr.dims.index(axis)

    values = arr.values
    for _ in range(order):
        as_arr = xr.DataArray(values, arr.coords, arr.dims)
        smoothed = smooth_fn(as_arr)
        values = np.gradient(smoothed.values, d_axis, axis=axis_idx)

    dn_arr = xr.DataArray(values, arr.coords, arr.dims, attrs=arr.attrs)

    if "id" in dn_arr.attrs:
        del dn_arr.attrs["id"]
        provenance(
            dn_arr,
            arr,
            {
                "what": "{}th derivative".format(order),
                "by": "dn_along_axis",
                "axis": axis,
                "order": order,
            },
        )

    return dn_arr


d2_along_axis = functools.partial(dn_along_axis, order=2)
d1_along_axis = functools.partial(dn_along_axis, order=1)



"""Utilities related to treating coordinates during data prep."""
import collections
import functools

import numpy as np


__all__ = ["disambiguate_coordinates"]


def disambiguate_coordinates(datasets, possibly_clashing_coordinates):
    """Finds and unifies duplicated coordinates or ambiguous coordinates.

    This is useful if two regions claim to have an energy axis, but one is a core level
    and so refers to a different energy range.
    """
    coords_set = collections.defaultdict(list)
    for d in datasets:
        for c in possibly_clashing_coordinates:
            if c in d.coords:
                coords_set[c].append(d.coords[c])

    conflicted = []
    for c in possibly_clashing_coordinates:
        different_coords = coords_set[c]
        if not different_coords:
            continue

        if not functools.reduce(
            lambda x, y: (np.array_equal(x[1], y) and x[0], y),
            different_coords,
            (True, different_coords[0]),
        )[0]:
            conflicted.append(c)

    after_deconflict = []
    for d in datasets:
        spectrum_name = list(d.data_vars.keys())[0]
        to_rename = {name: name + "-" + spectrum_name for name in d.dims if name in conflicted}
        after_deconflict.append(d.rename(to_rename))

    return after_deconflict



"""Infrastructure code for defining coordinate transforms and momentum conversion."""

import numpy as np

import xarray as xr
from typing import Any

__all__ = ["CoordinateConverter", "K_SPACE_BORDER", "MOMENTUM_BREAKPOINTS"]

K_SPACE_BORDER = 0.02
MOMENTUM_BREAKPOINTS = [0.0005, 0.001, 0.002, 0.005, 0.01, 0.02, 0.05, 0.1, 0.2, 0.5, 1]


class CoordinateConverter:
    """Infrastructure code to support a new coordinate conversion routine.

    In order to do coordinate conversion from c_i to d_i, we need to give functions
    c_i(d_j), i.e. to implement the inverse transform. This is so that we convert by
    interpolating the function from a regular grid of d_i values back to the original
    data expressed in c_i.

    From this, we can see what responsibilities these conversion classes hold:

    * They need to specify how to calculate c_i(d_j)
    * They need to cache computations so that computations of c_i(d_j) can be performed
      efficiently for different coordinates c_i
    * Because they know how to do the inverse conversion, they need to know how to choose
      reasonable grid bounds for the forward transform, so that this can be handled
      automatically.

    These different roles and how they are accomplished are discussed in detail below.
    """

    def __init__(
        self, ds: xr.Dataset, dim_order=None, calibration=None, *args, **kwargs
    ):
        """Intern the volume so that we can check on things during computation."""
        self.ds = ds
        self.dim_order = dim_order
        self.calibration = calibration

    def prep(self, ds: xr.Dataset):
        """Perform preprocessing of the dataset to convert before we start.

        The CoordinateConverter.prep method allows you to pre-compute some transformations
        that are common to the individual coordinate transform methods as an optimization.

        This is useful if you want the conversion methods to have separation of concern,
        but if it is advantageous for them to be able to share a computation of some
        variable. An example of this is in BE-kx-ky conversion, where computing k_p_tot
        is a step in both converting kx and ky, so we would like to avoid doing it twice.

        Of course, you can neglect this function entirely. Another technique is to simple
        cache computations as they arrive. This is the technique that is used in
        ConvertKxKy below
        """
        pass

    @property
    def is_slit_vertical(self) -> bool:
        """For hemispherical analyzers, whether the slit is vertical or horizontal.

        This is an ARPES specific detail, so this conversion code is not strictly general, but
        a future refactor could just push these details to a subclass.
        """
        # 89 - 91 degrees
        return np.abs(self.ds.S.lookup_offset_coord("alpha") - np.pi / 2) < (
            np.pi / 180
        )

    def kspace_to_BE(
        self, binding_energy: np.ndarray, *args: np.ndarray, **kwargs: Any
    ) -> np.ndarray:
        """The energy conservation equation for ARPES.

        This does not depend on any details of the angular conversion (it's the identity) so we can
        put the conversion code here in the base class.
        """
        return binding_energy

    def conversion_for(self, dim):
        """Fetches the method responsible for calculating `dim` from momentum coordinates."""
        pass

    def identity_transform(self, axis_name, *args, **kwargs):
        """Just returns the coordinate requested from args.

        Useful if the transform is the identity.
        """
        return args[self.dim_order.index(axis_name)]

    def get_coordinates(self, resolution: dict = None, bounds: dict = None):
        """Calculates the coordinates which should be used in momentum space."""
        coordinates = {}
        coordinates["eV"] = self.ds.coords["eV"]
        return coordinates



"""Extends lmfit to support curve fitting on xarray instances."""
import operator
import warnings
from lmfit.models import GaussianModel
import xarray as xr
import lmfit as lf
import numpy as np

__all__ = ["XModelMixin", "gaussian_convolve"]


def dict_to_parameters(dict_of_parameters) -> lf.Parameters:
    params = lf.Parameters()

    for param_name, param in dict_of_parameters.items():
        params[param_name] = lf.Parameter(param_name, **param)

    return params


class XModelMixin(lf.Model):
    """A mixin providing curve fitting for ``xarray.DataArray`` instances.

    This amounts mostly to making `lmfit` coordinate aware, and providing
    a translation layer between xarray and raw np.ndarray instances.

    Subclassing this mixin as well as an lmfit Model class should bootstrap
    an lmfit Model to one that works transparently on xarray data.

    Alternatively, you can use this as a model base in order to build new models.

    The core method here is `guess_fit` which is a convenient utility that performs both
    a `lmfit.Model.guess`, if available, before populating parameters and
    performing a curve fit.

    __add__ and __mul__ are also implemented, to ensure that the composite model
    remains an instance of a subclass of this mixin.
    """

    n_dims = 1
    dimension_order = None

    def guess_fit(
        self,
        data,
        params=None,
        weights=None,
        guess=True,
        debug=False,
        prefix_params=True,
        transpose=False,
        **kwargs
    ):
        """Performs a fit on xarray data after guessing parameters.

        Params allows you to pass in hints as to what the values and bounds on parameters
        should be. Look at the lmfit docs to get hints about structure
        """
        if params is not None and not isinstance(params, lf.Parameters):
            params = dict_to_parameters(params)

        if transpose:
            assert (
                len(data.dims) == 1
                and "You cannot transpose (invert) a multidimensional array (scalar field)."
            )

        coord_values = {}
        if "x" in kwargs:
            coord_values["x"] = kwargs.pop("x")

        real_data, flat_data = data, data

        new_dim_order = None
        if isinstance(data, xr.DataArray):
            real_data, flat_data = data.values, data.values
            assert len(real_data.shape) == self.n_dims

            if self.n_dims == 1:
                coord_values["x"] = data.coords[list(data.indexes)[0]].values
            else:

                def find_appropriate_dimension(dim_or_dim_list):
                    if isinstance(dim_or_dim_list, str):
                        assert dim_or_dim_list in data.dims
                        return dim_or_dim_list

                    else:
                        intersect = set(dim_or_dim_list).intersection(data.dims)
                        assert len(intersect) == 1
                        return list(intersect)[0]

                # resolve multidimensional parameters
                if self.dimension_order is None or all(d is None for d in self.dimension_order):
                    new_dim_order = data.dims
                else:
                    new_dim_order = [
                        find_appropriate_dimension(dim_options)
                        for dim_options in self.dimension_order
                    ]

                if list(new_dim_order) != list(data.dims):
                    warnings.warn("Transposing data for multidimensional fit.")
                    data = data.transpose(*new_dim_order)

                coord_values = {k: v.values for k, v in data.coords.items() if k in new_dim_order}
                real_data, flat_data = data.values, data.values.ravel()

        real_weights = weights
        if isinstance(weights, xr.DataArray):
            if self.n_dims == 1:
                real_weights = real_weights.values
            else:
                if new_dim_order is not None:
                    real_weights = weights.transpose(*new_dim_order).values.ravel()
                else:
                    real_weights = weights.values.ravel()

        if transpose:
            cached_coordinate = list(coord_values.values())[0]
            coord_values[list(coord_values.keys())[0]] = real_data
            real_data = cached_coordinate
            flat_data = real_data

        if guess:
            guessed_params = self.guess(real_data, **coord_values)
        else:
            guessed_params = self.make_params()

        if params is not None:
            for k, v in params.items():
                if isinstance(v, dict):
                    if prefix_params:
                        guessed_params[self.prefix + k].set(**v)
                    else:
                        guessed_params[k].set(**v)

            guessed_params.update(params)

        result = None
        try:
            result = super().fit(
                flat_data, guessed_params, **coord_values, weights=real_weights, **kwargs
            )
            result.independent = coord_values
            result.independent_order = new_dim_order
        except Exception as e:
            print(e)
            if debug:
                import pdb

                pdb.post_mortem(e.__traceback__)
        finally:
            return result

    def xguess(self, data, **kwargs):
        """Tries to determine a guess for the parameters."""
        x = kwargs.pop("x", None)

        real_data = data
        if isinstance(data, xr.DataArray):
            real_data = data.values
            assert len(real_data.shape) == 1
            x = data.coords[list(data.indexes)[0]].values

        return self.guess(real_data, x=x, **kwargs)

    def __add__(self, other):
        """Implements `+`."""
        comp = XAdditiveCompositeModel(self, other, operator.add)

        assert self.n_dims == other.n_dims
        comp.n_dims = other.n_dims

        return comp

    def __mul__(self, other):
        """Implements `*`."""
        comp = XMultiplicativeCompositeModel(self, other, operator.mul)

        assert self.n_dims == other.n_dims
        comp.n_dims = other.n_dims

        return comp


class XAdditiveCompositeModel(lf.CompositeModel, XModelMixin):
    """xarray coordinate aware composite model corresponding to the sum of two models."""

    def guess(self, data, x=None, **kwargs):
        pars = self.make_params()
        guessed = {}
        for c in self.components:
            guessed.update(c.guess(data, x=x, **kwargs))

        for k, v in guessed.items():
            pars[k] = v

        return pars


class XMultiplicativeCompositeModel(lf.CompositeModel, XModelMixin):
    """xarray coordinate aware composite model corresponding to the sum of two models.

    Currently this just copies ``+``, might want to adjust things!
    """

    def guess(self, data, x=None, **kwargs):
        pars = self.make_params()
        guessed = {}
        for c in self.components:
            guessed.update(c.guess(data, x=x, **kwargs))

        for k, v in guessed.items():
            pars[k] = v

        return pars


class XConvolutionCompositeModel(lf.CompositeModel, XModelMixin):
    """Work in progress for convolving two ``Model``."""

    def guess(self, data, x=None, **kwargs):
        pars = self.make_params()
        guessed = {}

        for c in self.components:
            if c.prefix == "conv_":
                # don't guess on the convolution term
                continue

            guessed.update(c.guess(data, x=x, **kwargs))

        for k, v in guessed.items():
            pars[k] = v

        return pars


def gaussian_convolve(model_instance):
    """Produces a model that consists of convolution with a Gaussian kernel."""
    return XConvolutionCompositeModel(model_instance, GaussianModel(prefix="conv_"), np.convolve)



"""Contains self-energy analysis routines."""
import xarray as xr
import lmfit as lf
import numpy as np

from arpes.constants import METERS_PER_SECOND_PER_EV_ANGSTROM, HBAR_PER_EV

from arpes.fits.utilities import broadcast_model
from arpes.fits.fit_models import AffineBackgroundModel, LorentzianModel, LinearModel

from typing import Optional, Union

__all__ = (
    "to_self_energy",
    "fit_for_self_energy",
    "estimate_bare_band",
    "quasiparticle_lifetime",
)


BareBandType = Union[xr.DataArray, str, lf.model.ModelResult]
DispersionType = Union[xr.DataArray, xr.Dataset]


def get_peak_parameter(data: xr.DataArray, parameter_name: str) -> xr.DataArray:
    """Extracts a parameter from a potentially prefixed peak-like component.

    Works so long as there is only a single peak defined in the model.

    Args:
        data: The input data containing the peak fitting results.
        parameter_name: The name of the parameter which should be extracted.

    Returns:
        The array of parameters corresponding to a peak in a single-peak curve fit.
    """
    first_item = data.values.ravel()[0]
    peak_like = (
        lf.models.LorentzianModel,
        lf.models.VoigtModel,
        lf.models.GaussianModel,
        lf.models.PseudoVoigtModel,
    )

    if isinstance(first_item, lf.model.ModelResult):
        if isinstance(first_item.model, lf.model.CompositeModel):
            peak_like_components = [
                c for c in first_item.model.components if isinstance(c, peak_like)
            ]
            assert len(peak_like_components) == 1

            return data.F.p("{}{}".format(peak_like_components[0].prefix, parameter_name))
        else:
            return data.F.p(parameter_name)

    raise ValueError(
        "Unsupported dispersion {}, expected xr.DataArray[lmfit.ModelResult]".format(type(data))
    )


def local_fermi_velocity(bare_band: xr.DataArray):
    """Calculates the band velocity under assumptions of a linear bare band."""
    fitted_model = LinearModel().guess_fit(bare_band)
    raw_velocity = fitted_model.params["slope"].value

    if "eV" in bare_band.dims:
        # the "y" values are in `bare_band` are momenta and the "x" values are energy, therefore
        # the slope is dy/dx = dk/dE
        raw_velocity = 1 / raw_velocity

    return raw_velocity * METERS_PER_SECOND_PER_EV_ANGSTROM


def estimate_bare_band(dispersion: xr.DataArray, bare_band_specification: Optional[str] = None):
    """Estimates the bare band from a fitted dispersion.

    This can be done in a few ways:
    #. None: Equivalent to 'baseline_linear' below
    #. `'linear'`: A linear fit to the dispersion is used, and this also provides the fermi_velocity
    #. `'ransac_linear'`: A linear fit with random sample consensus (RANSAC) region will be used and this
      also provides the `fermi_velocity`
    #. `'hough'`: Hough transform based method

    Args:
        dispersion: The array of the fitted peak locations.
        bare_band_specification: What kind of bare band to assume. One of "linear",
          "ransac_linear", and "hough".

    Returns:
        An estimate of the bare band dispersion.
    """
    try:
        centers = get_peak_parameter(dispersion, "center")
    except ValueError:
        centers = dispersion

    mom_options = [d for d in dispersion.dims if d in {"k", "kp", "kx", "ky", "kz"}]
    assert len(mom_options) <= 1
    fit_dimension = "eV" if "eV" in dispersion.dims else mom_options[0]

    if bare_band_specification is None:
        bare_band_specification = "ransac_linear"

    initial_linear_fit = LinearModel().guess_fit(centers)
    if bare_band_specification == "linear":
        fitted_model = initial_linear_fit
    elif bare_band_specification == "ransac_linear":
        from skimage.measure import LineModelND, ransac

        min_samples = len(centers.coords[fit_dimension]) // 10
        residual_threshold = np.median(np.abs(initial_linear_fit.residual)) * 1
        ransac_model, inliers = ransac(
            np.stack([centers.coords[fit_dimension], centers]).T,
            LineModelND,
            max_trials=1000,
            min_samples=min_samples,
            residual_threshold=residual_threshold,
        )
        inlier_data = centers.where(
            xr.DataArray(
                inliers,
                coords=dict([[fit_dimension, centers.coords[fit_dimension]]]),
                dims=[fit_dimension],
            ),
            drop=True,
        )

        fitted_model = LinearModel().guess_fit(inlier_data)
    elif bare_band_specification == "hough":
        raise NotImplementedError("Hough Transform estimate of bare band not yet supported.")
    else:
        raise ValueError("Unrecognized bare band type: {}".format(bare_band_specification))

    ys = fitted_model.eval(x=centers.coords[fit_dimension])
    return xr.DataArray(ys, centers.coords, centers.dims)


def quasiparticle_lifetime(self_energy: xr.DataArray, bare_band: xr.DataArray) -> xr.DataArray:
    """Calculates the quasiparticle mean free path in meters (meters!).

    The bare band is used to calculate the band/Fermi velocity
    and internally the procedure to calculate the quasiparticle lifetime is used.

    Args:
        self_energy: The measured or estimated self-energy.
        bare_band: The bare band defining the band velocity.

    Returns:
        An estimate of the quasiparticle lifetime along the band.
    """
    imaginary_part = np.abs(np.imag(self_energy)) / 2
    return HBAR_PER_EV / imaginary_part


def quasiparticle_mean_free_path(
    self_energy: xr.DataArray, bare_band: xr.DataArray
) -> xr.DataArray:
    lifetime = quasiparticle_lifetime(self_energy, bare_band)
    return lifetime * local_fermi_velocity(bare_band)


def to_self_energy(
    dispersion: xr.DataArray,
    bare_band: Optional[BareBandType] = None,
    k_independent=True,
    fermi_velocity=None,
) -> xr.Dataset:
    r"""Converts MDC fit results into the self energy.

    This largely consists of extracting
    out the linewidth and the difference between the dispersion and the bare band value.

    lorentzian(x, amplitude, center, sigma) =
        (amplitude / pi) * sigma/(sigma^2 + ((x-center))**2)

    Once we have the curve-fitted dispersion we can calculate the self energy if we also
    know the bare-band dispersion. If the bare band is not known, then at least the imaginary
    part of the self energy is still calculable, and a guess as to the real part can be
    calculated under assumptions of the bare band dispersion as being free electron like wih
    effective mass m* or being Dirac like (these are equivalent at low enough energy).

    Acceptabe bare band spefications are discussed in detail in `estimate_bare_band` above.

    To future readers of the code, please note that the half-width half-max of a Lorentzian is equal to the
    $\gamma$ parameter, which defines the imaginary part of the self energy.

    Args:
        dispersion
        bare_band
        k_independent
        fermi_velocity

    Returns:
        The equivalent self energy from the bare band and the measured dispersion.
    """
    if not k_independent:
        raise NotImplementedError(
            "PyARPES does not currently support self energy analysis "
            "except in the k-independent formalism."
        )

    if isinstance(dispersion, xr.Dataset):
        dispersion = dispersion.results

    from_mdcs = "eV" in dispersion.dims  # if eV is in the dimensions, then we fitted MDCs
    estimated_bare_band = estimate_bare_band(dispersion, bare_band)

    if fermi_velocity is None:
        fermi_velocity = local_fermi_velocity(estimated_bare_band)

    imaginary_part = get_peak_parameter(dispersion, "fwhm") / 2
    centers = get_peak_parameter(dispersion, "center")

    if from_mdcs:
        imaginary_part *= fermi_velocity / METERS_PER_SECOND_PER_EV_ANGSTROM
        real_part = -(
            (centers * fermi_velocity / METERS_PER_SECOND_PER_EV_ANGSTROM)
            - dispersion.coords["eV"].values
        )
    else:
        real_part = centers - bare_band

    self_energy = xr.DataArray(
        real_part + 1.0j * imaginary_part,
        coords=dispersion.coords,
        dims=dispersion.dims,
    )

    return xr.Dataset({"self_energy": self_energy, "bare_band": estimated_bare_band})


def fit_for_self_energy(
    data: xr.DataArray, method="mdc", bare_band: Optional[BareBandType] = None, **kwargs
) -> xr.Dataset:
    """Fits for the self energy of a dataset containing a single band.

    Args:
        data: The input data.
        method: one of 'mdc' and 'edc'
        bare_band: Optionally, the bare band. If None is provided the bare band will be estimated.

    Returns:
        The self energy resulting from curve-fitting.
    """
    if method == "mdc":
        fit_results = broadcast_model(
            [LorentzianModel, AffineBackgroundModel], data, "eV", **kwargs
        )
    else:
        possible_mometum_dims = ("phi", "theta", "psi", "beta", "kp", "kx", "ky", "kz")
        mom_axes = set(data.dims).intersection(possible_mometum_dims)

        if len(mom_axes) > 1:
            raise ValueError("Too many possible momentum dimensions, please clarify.")
        fit_results = broadcast_model(
            [LorentzianModel, AffineBackgroundModel], data, list(mom_axes)[0], **kwargs
        )

    return to_self_energy(fit_results, bare_band=bare_band)



"""Contains very basic spin-ARPES analysis routines."""
import xarray as xr
from arpes.provenance import update_provenance
from arpes.typing import DataType
from arpes.utilities import normalize_to_dataset
from arpes.utilities.math import polarization

__all__ = (
    "to_intensity_polarization",
    "to_up_down",
    "normalize_sarpes_photocurrent",
)


@update_provenance("Normalize SARPES by photocurrent")
def normalize_sarpes_photocurrent(data: DataType) -> DataType:
    """Normalizes the down channel so that it matches the up channel in terms of mean photocurrent.

    Destroys the integrity of "count" data because we have scaled individual arrivals.

    Args:
        data: The input data which does not need to consist of count data.

    Returns:
        Scaled data. Independently, photocurrent up and down channels are used to perform scaling.
    """
    copied = data.copy(deep=True)
    copied.down.values = (copied.down * (copied.photocurrent_up / copied.photocurrent_down)).values
    return copied


@update_provenance("Convert polarization data to up-down spin channels")
def to_up_down(data: DataType) -> xr.Dataset:
    """Converts from [intensity, polarization] representation to [up, down] representation.

    This is the inverse function to `to_intensity_polarization`, neglecting the role of the
    sherman function.

    Args:
        data: The input data

    Returns:
        The data after conversion to up-down representation.
    """
    assert "intensity" in data.data_vars and "polarization" in data.data_vars

    return xr.Dataset(
        {
            "up": data.intensity * (1 + data.polarization),
            "down": data.intensity * (1 - data.polarization),
        }
    )


@update_provenance("Convert up-down spin channels to polarization")
def to_intensity_polarization(
    data: DataType, perform_sherman_correction: bool = False
) -> xr.Dataset:
    """Converts from [up, down] representation to [intensity, polarization] representation.

    This is the inverse function to `to_up_down`.

    In this future, we should also make this also work with the timing signals.

    Args:
        data: The input data

    Returns:
        The data after conversion to intensity-polarization representation.
    """
    data = normalize_to_dataset(data)

    assert "up" in data.data_vars and "down" in data.data_vars

    intensity = data.up + data.down
    spectrum_polarization = polarization(data.up, data.down)

    sherman_correction = 1.0
    if perform_sherman_correction:
        sherman_correction = data.S.sherman_function

    return xr.Dataset(
        {"intensity": intensity, "polarization": spectrum_polarization / sherman_correction}
    )



"""Defines a widget which provides a 1D browsable `lmfit.model.ModelResult`."""

from PyQt5 import QtCore, QtWidgets
from arpes.utilities.qt.data_array_image_view import DataArrayPlot
from typing import Optional
import pyqtgraph as pg
import xarray as xr
import numpy as np
import math
from PyQt5.QtWidgets import QLayout, QGridLayout, QWidget
import lmfit
from weakref import ReferenceType

from arpes.utilities.qt import qt_info

__all__ = ["FitInspectionPlot"]


class LabelParametersInfoView(QtWidgets.QLabel):
    def __init__(self, parent=None):
        """Nothing interesting to do here, we just delegate to super."""
        super().__init__(parent)
        self.setText("")

    def set_model_result(self, model_result: lmfit.model.ModelResult):
        """Converts the ModelResult to the HTML representation and sets page contents."""
        assert model_result is not None
        self.setText(model_result._repr_multiline_text_(short=True))


class FitInspectionPlot(QWidget):
    """Implements the `ModeResult` inspection tool."""

    layout: QLayout = None
    result: Optional[lmfit.model.ModelResult] = None
    root: ReferenceType = None

    def x(self) -> np.ndarray:
        """Returns the single fit coordinate along the model data."""
        fit_dim = self.root().dataset.F.fit_dimensions[0]
        return self.root().dataset.coords[fit_dim].values

    @property
    def data(self) -> np.ndarray:
        """Returns the values of the data used for fitting."""
        return self.result.data

    @property
    def residual(self) -> np.ndarray:
        """Returns the residual from the fit."""
        return self.result.residual

    @property
    def eval_model(self) -> np.ndarray:
        """Returns the values of the fit at the x coordinates.

        Rather than accomplishing this by actually calling the model
        eval, we cheat by using the original data and the residual to back out
        the fit data.
        """
        return self.data - self.residual

    @property
    def init_eval_model(self) -> np.ndarray:
        """Returns the initial values of the fit.

        This is useful to see coarsely whether the initial parameter
        values being used are leading to reasonable and convergent fits.
        """
        return self.result.init_fit

    def set_model_result(self, model_result: lmfit.model.ModelResult) -> None:
        """Updates the current widgets according to the new coords/model result.

        In order to do this, we need to update the three lineplots on the plot widget:
        * The data
        * The residual
        * The evaled data
        * (TODO, the component data)

        and we need to update the model info section.
        """
        # assign the model result onto this instance
        self.result = model_result

        # update the plotted data
        x = self.x()
        coords = {"x": x}
        dims = ["x"]

        p1 = self.inner_plot.plot(xr.DataArray(self.data, coords, dims), clear=True)
        p2 = self.inner_plot.plot(xr.DataArray(self.residual, coords, dims))
        p3 = self.inner_plot.plot(xr.DataArray(self.eval_model, coords, dims))
        p4 = self.inner_plot.plot(xr.DataArray(self.init_eval_model, coords, dims))

        plot_width = int(math.ceil(qt_info.inches_to_px(0.02)))
        p1.setPen(pg.mkPen(width=plot_width, color=(0, 0, 0)))
        p2.setPen(pg.mkPen(width=plot_width, color=(255, 0, 0)))
        p3.setPen(pg.mkPen(width=plot_width, color=(50, 200, 20), style=QtCore.Qt.DotLine))
        p4.setPen(pg.mkPen(width=plot_width, color=(70, 150, 70), style=QtCore.Qt.DotLine))

        # update the model info
        self.model_info.set_model_result(model_result)

    def close(self):
        """Clean up references so we do not cause GC issues and Qt crashes."""
        self.model_info.close()
        self.inner_plot.close()
        self.model_info = None
        self.inner_plot = None
        self.root = None
        self.result = None

        super().close()

    def __init__(self, root, orientation, name=None, *args, **kwargs):
        """Performs initial registration of the widgets and sets up layout."""
        super().__init__(*args, **kwargs)
        self.layout = QGridLayout()
        self.model_info = LabelParametersInfoView()
        self.inner_plot = DataArrayPlot(root, orientation, name=name)

        self.setLayout(self.layout)
        self.layout.addWidget(self.inner_plot, 0, 0)
        self.layout.addWidget(self.model_info, 0, 1)

        self.root = root



"""Provides xarray aware pyqtgraph plotting widgets."""

# pylint: disable=import-error
import pyqtgraph as pg
import numpy as np
from scipy import interpolate
from weakref import ReferenceType

from .utils import PlotOrientation

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from .app import SimpleApp

__all__ = (
    "DataArrayImageView",
    "DataArrayPlot",
)


class CoordAxis(pg.AxisItem):
    def __init__(self, dim_index, *args, **kwargs):
        self.dim_index = dim_index
        self.coord = None
        self.interp = None
        super().__init__(*args, **kwargs)

    def setImage(self, image):
        self.coord = image.coords[image.dims[self.dim_index]].values
        self.interp = interpolate.interp1d(
            np.arange(0, len(self.coord)), self.coord, fill_value="extrapolate"
        )

    def tickStrings(self, values, scale, spacing):
        try:
            return ["{:.3f}".format(f) for f in self.interp(values)]
        except TypeError:
            return super().tickStrings(values, scale, spacing)


class DataArrayPlot(pg.PlotWidget):
    """A plot for 1D xr.DataArray instances with a coordinate aware axis."""

    def __init__(self, root: ReferenceType, orientation, *args, **kwargs):
        """
        Use custom axes so that we can provide coordinate-ful rather than pixel based
        values.
        """
        self._root = root
        self.orientation = orientation

        axis_or = "bottom" if orientation == PlotOrientation.Horizontal else "left"
        self._coord_axis = CoordAxis(dim_index=0, orientation=axis_or)

        super().__init__(axisItems=dict([[axis_or, self._coord_axis]]), *args, **kwargs)

    @property
    def root(self) -> "SimpleApp":
        """Unpacks the weak reference to the root object."""
        return self._root()

    def plot(self, data, *args, **kwargs):
        """Updates the UI with new data.

        Data also needs to be forwarded to the coordinate axis in case of transpose
        or changed range of data.
        """
        y = data.values
        self._coord_axis.setImage(data)

        if self.orientation == PlotOrientation.Horizontal:
            return self.plotItem.plot(
                np.arange(0, len(y)),
                y,
                pen=pg.mkPen(color=(68, 1, 84), width=3),
                *args,
                **kwargs
            )
        else:
            return self.plotItem.plot(
                y,
                np.arange(0, len(y)),
                pen=pg.mkPen(color=(68, 1, 84), width=3),
                *args,
                **kwargs
            )


class DataArrayImageView(pg.ImageView):
    """
    ImageView that transparently handles xarray data, including setting axis and
    coordinate information. This makes it easier to build interactive applications
    around realistic scientific datasets.
    """

    def __init__(self, root: ReferenceType, *args, **kwargs):
        """
        Use custom axes so that we can provide coordinate-ful rather than pixel based
        values.
        """
        self._root = root
        self._coord_axes = {
            "left": CoordAxis(dim_index=1, orientation="left"),
            "bottom": CoordAxis(dim_index=0, orientation="bottom"),
        }
        plot_item = pg.PlotItem(axisItems=self._coord_axes)
        self.plot_item = plot_item
        super().__init__(view=plot_item, *args, **kwargs)

        self.view.invertY(False)

    @property
    def root(self) -> "SimpleApp":
        """Unpacks the weak reference to the root object."""
        return self._root()

    def setImage(self, img, keep_levels=False, *args, **kwargs):
        """Accepts an xarray.DataArray instead of a numpy array."""
        if keep_levels:
            levels = self.getLevels()

        for axis in self._coord_axes.values():
            axis.setImage(img)

        super().setImage(img.values, *args, **kwargs)

        if keep_levels:
            self.setLevels(*levels)

    def recompute(self):
        """A hook to recompute UI state, not used by this widget."""
        pass



"""A help dialog showing keyboard shortcuts for Qt application."""

# pylint: disable=import-error

from PyQt5 import QtCore, QtWidgets

from arpes.utilities.qt.ui import PRETTY_KEYS, label, vertical

__all__ = ("BasicHelpDialog",)


class BasicHelpDialog(QtWidgets.QDialog):
    """A help dialog showing keyboard shortcuts for Qt application."""

    def __init__(self, shortcuts=None):
        """Initialize the help window and build widgets for the registered shortcuts."""
        super().__init__()

        if shortcuts is None:
            shortcuts = []

        self.layout = QtWidgets.QVBoxLayout()

        keyboard_shortcuts_info = QtWidgets.QGroupBox(title="Keyboard Shortcuts")
        keyboard_shortcuts_layout = QtWidgets.QGridLayout()
        for i, shortcut in enumerate(shortcuts):
            keyboard_shortcuts_layout.addWidget(
                label(", ".join(PRETTY_KEYS[k] for k in shortcut.chord), wordWrap=True),
                i,
                0,
            )
            keyboard_shortcuts_layout.addWidget(label(shortcut.label), i, 1)

        keyboard_shortcuts_info.setLayout(keyboard_shortcuts_layout)

        aboutInfo = QtWidgets.QGroupBox(title="About")
        aboutLayout = vertical(
            label(
                "QtTool is the work of Conrad Stansbury, with much inspiration "
                "and thanks to the authors of ImageTool. QtTool is distributed "
                "as part of the PyARPES data analysis framework.",
                wordWrap=True,
            ),
            label(
                "Complaints and feature requests should be directed to chstan@berkeley.edu.",
                wordWrap=True,
            ),
        )

        from arpes.utilities.qt import qt_info  # circular dependency

        aboutInfo.setFixedHeight(qt_info.inches_to_px(1))

        self.layout.addWidget(keyboard_shortcuts_info)
        self.layout.addWidget(aboutInfo)
        self.setLayout(self.layout)

        self.setWindowTitle(f"Interactive Utility Help")
        self.setFixedSize(*qt_info.inches_to_px([2, 4]))

    def keyPressEvent(self, event):
        """If the user presed H we should toggle the dialog, or close it if they pressed Esc."""
        if event.key() == QtCore.Qt.Key_H or event.key() == QtCore.Qt.Key_Escape:
            self._main_window._help_dialog = None  # pylint: disable=protected-access
            self.close()



"""Implements forward and reverse trapezoidal corrections."""

import warnings
import numpy as np
import xarray as xr

import numba

from typing import Any, Callable

from arpes.utilities import normalize_to_spectrum

from .base import CoordinateConverter
from .core import convert_coordinates

__all__ = ["apply_trapezoidal_correction"]


@numba.njit(parallel=True)
def _phi_to_phi(energy, phi, phi_out, l_fermi, l_volt, r_fermi, r_volt):
    """Performs reverse coordinate interpolation using four angular waypoints.

    Args:
        energy: The binding energy in the corrected coordinate space
        phi: The angle in the corrected coordinate space
        phi_out: The array to populate with the measured phi angles
        l_fermi: The measured phi coordinate of the left edge of the hemisphere's range
           at the Fermi level
        l_volt: The measured phi coordinate of the left edge of the hemisphere's range
           at a binding energy of 1 eV (eV = -1.0)
        r_fermi: The measured phi coordinate of the right edge of the hemisphere's range
           at the Fermi level
        r_volt: The measured phi coordinate of the right edge of the hemisphere's range
           at a binding energy of 1 eV (eV = -1.0)
    """
    for i in numba.prange(len(phi)):
        l = l_fermi - energy[i] * (l_volt - l_fermi)
        r = r_fermi - energy[i] * (r_volt - r_fermi)

        # These are the forward equations, we can just invert them below
        # c = (phi[i] - l) / (r - l)
        # phi_out[i] = l_fermi + c * (r_fermi - l_fermi)

        dac_da = (r - l) / (r_fermi - l_fermi)
        phi_out[i] = (phi[i] - l_fermi) * dac_da + l


@numba.njit(parallel=True)
def _phi_to_phi_forward(energy, phi, phi_out, l_fermi, l_volt, r_fermi, r_volt):
    """The inverse transform to ``_phi_to_phi``. See that function for details."""
    for i in numba.prange(len(phi)):
        l = l_fermi - energy[i] * (l_volt - l_fermi)
        r = r_fermi - energy[i] * (r_volt - r_fermi)

        # These are the forward equations
        c = (phi[i] - l) / (r - l)
        phi_out[i] = l_fermi + c * (r_fermi - l_fermi)


class ConvertTrapezoidalCorrection(CoordinateConverter):
    """A converter for applying the trapezoidal correction to ARPES data."""

    def __init__(self, *args: Any, corners: list[dict[str, float]], **kwargs: Any):
        super().__init__(*args, **kwargs)
        self.phi = None

        # we normalize the corners so that they are equivalent to four corners at the Fermi level
        # and one volt below.
        c1, c2, c3, c4 = sorted(corners, key=lambda x: x["phi"])
        c1, c2 = sorted([c1, c2], key=lambda x: x["eV"])
        c3, c4 = sorted([c3, c4], key=lambda x: x["eV"])

        # now, corners are in
        # (c1, c2, c3, c4) = (LL, UL, LR, UR) order

        left_per_volt = (c1["phi"] - c2["phi"]) / (c1["eV"] - c2["eV"])
        left_phi_fermi = c2["phi"] - c2["eV"] * left_per_volt
        left_phi_one_volt = left_phi_fermi - left_per_volt

        right_per_volt = (c3["phi"] - c4["phi"]) / (c3["eV"] - c4["eV"])
        right_phi_fermi = c3["phi"] - c4["eV"] * right_per_volt
        right_phi_one_volt = right_phi_fermi - right_per_volt

        self.corner_angles = (
            left_phi_fermi,
            left_phi_one_volt,
            right_phi_fermi,
            right_phi_one_volt,
        )

    def get_coordinates(self, *args, **kwargs):
        return self.ds.indexes

    def conversion_for(self, dim: str) -> Callable:
        def with_identity(*args, **kwargs):
            return self.identity_transform(dim, *args, **kwargs)

        return {
            "phi": self.phi_to_phi,
        }.get(dim, with_identity)

    def phi_to_phi(
        self, binding_energy: np.ndarray, phi: np.ndarray, *args: Any, **kwargs: Any
    ):
        if self.phi is not None:
            return self.phi
        self.phi = np.zeros_like(phi)
        _phi_to_phi(binding_energy, phi, self.phi, *self.corner_angles)
        return self.phi

    def phi_to_phi_forward(
        self, binding_energy: np.ndarray, phi: np.ndarray, *args: Any, **kwargs: Any
    ):
        phi_out = np.zeros_like(phi)
        _phi_to_phi_forward(binding_energy, phi, phi_out, *self.corner_angles)
        return phi_out


def apply_trapezoidal_correction(
    data: xr.DataArray,
    corners: list[dict[str, float]],
) -> xr.DataArray:
    """Applies the trapezoidal correction to data in angular units by linearly interpolating slices.

    Shares some code with standard coordinate conversion, i.e. to momentum, because you can think of
    this as performing a coordinate conversion between two angular coordinate sets, the measured angles
    and the true angles.

    Args:
        data: The xarray instances to perform correction on
        corners: These don't actually have to be corners, but are waypoints of the conversion. Use points near the Fermi
            level and near the bottom of the spectrum just at the edge of recorded angular region.


    Returns:
        The corrected data.
    """

    if isinstance(data, dict):
        warnings.warn(
            "Treating dict-like data as an attempt to forward convert a single coordinate."
        )
        converter = ConvertTrapezoidalCorrection(None, [], corners=corners)
        result = dict(data)
        result["phi"] = converter.phi_to_phi_forward(
            np.array([data["eV"]]), np.array([data["phi"]])
        )[0]
        return result

    if isinstance(data, xr.Dataset):
        warnings.warn(
            "Remember to use a DataArray not a Dataset, attempting to extract spectrum and copy attributes."
        )
        attrs = data.attrs.copy()
        data = normalize_to_spectrum(data)
        data.attrs.update(attrs)

    original_coords = data.coords

    if "phi" not in data.dims:
        raise ValueError("The data must have a phi coordinate.")
    removed = [d for d in data.dims if d not in ["eV", "phi"]]
    data = data.transpose(*(["eV", "phi"] + removed))
    converted_dims = data.dims

    restore_index_like_coordinates = {r: data.coords[r].values for r in removed}
    new_index_like_coordinates = {
        r: np.arange(len(data.coords[r].values)) for r in removed
    }
    data = data.assign_coords(**new_index_like_coordinates)

    converter = ConvertTrapezoidalCorrection(data, converted_dims, corners=corners)
    converted_coordinates = converter.get_coordinates()

    result = convert_coordinates(
        data,
        converted_coordinates,
        {
            "dims": data.dims,
            "transforms": dict(
                zip(data.dims, [converter.conversion_for(d) for d in data.dims])
            ),
        },
    )

    result = result.assign_coords(**restore_index_like_coordinates)
    result = result.assign_coords(
        **{c: v for c, v in original_coords.items() if c not in result.coords}
    )
    result = result.assign_attrs(data.attrs)
    return result



import os
from pathlib import Path
import numpy as np
import xarray as xr
from matplotlib import pyplot as plt

from arpes.io import load_data, load_folder, export_dataset
from arpes.plotting.stack_plot import stack_dispersion_plot
from arpes.utilities.conversion import convert_to_kspace
from arpes.corrections import fix_fermi_edge
from arpes.analysis.general import rebin, normalize
from arpes.analysis.filters import gaussian_filter_arr
from arpes import xarray_extensions

from arpes.endstations.base import add_endstation
from arpes.endstations.plugins import (
    MAESTROMicroARPESEndstation,
    SPEEMEndstation,
    ALGMainChamber,
    HHGEndstation,
)

for endstation in [
    MAESTROMicroARPESEndstation,
    SPEEMEndstation,
    ALGMainChamber,
    HHGEndstation,
]:
    add_endstation(endstation)
xr.set_options(keep_attrs=True)
plt.rcParams["image.cmap"] = "magma"

cwd = Path(os.getcwd())
# We allow for one level of nesting in the directory structure
# Nominally, the only subdirectories should specify the measurement date
if cwd.name == "notebooks":
    root = cwd.parent
    data_root = root / "data"
    exports_root = data_root / "exports"
    results_root = root / "results"
else:
    if cwd.parent.name != "notebooks":
        from warnings import warn as _warn

        _warn(
            "The designated directory structure isn't being met. "
            "The provided paths may not be correct."
        )

    measurement_date = cwd.name
    root = cwd.parent.parent
    data_root = root / "data" / measurement_date
    exports_root = data_root / "exports"
    results_root = root / "results"



"""Utilities related to treatment of coordinate axes."""

import copy
import functools
from typing import Container

import numpy as np
from scipy.ndimage import geometric_transform

import xarray as xr
from arpes.provenance import provenance, update_provenance
from arpes.typing import DataType
from arpes.utilities import lift_dataarray_to_generic
from arpes.utilities.normalize_datatype import normalize_to_spectrum

__all__ = (
    "flip_axis",
    "normalize_dim",
    "dim_normalizer",
    "transform_dataarray_axis",
    "normalize_total",
    "sort_axis",
)


@update_provenance("Sort Axis")
def sort_axis(data: xr.DataArray, axis_name):
    """Sorts slices of `data` along `axis_name` so that they lie in order."""
    assert isinstance(data, xr.DataArray)
    copied = data.copy(deep=True)
    coord = data.coords[axis_name].values
    order = np.argsort(coord)

    copied.values = np.take(copied.values, order, axis=list(data.dims).index(axis_name))
    copied.coords[axis_name] = np.sort(copied.coords[axis_name])
    return copied


@update_provenance("Flip data along axis")
def flip_axis(arr: xr.DataArray, axis_name, flip_data=True):
    """Flips the coordinate values along an axis without changing the data as well."""
    coords = copy.deepcopy(arr.coords)
    coords[axis_name] = coords[axis_name][::-1]

    return xr.DataArray(
        np.flip(arr.values, arr.dims.index(axis_name)) if flip_data else arr.values,
        coords,
        arr.dims,
        attrs=arr.attrs,
    )


def soft_normalize_dim(arr: xr.DataArray, dim_or_dims, keep_id=False, amp_limit=100):
    dims = dim_or_dims
    if isinstance(dim_or_dims, str):
        dims = [dims]

    summed_arr = arr.fillna(arr.mean()).sum([d for d in arr.dims if d not in dims])
    normalized_arr = arr / (summed_arr / np.prod(summed_arr.shape))

    to_return = xr.DataArray(
        normalized_arr.values, arr.coords, arr.dims, attrs=arr.attrs
    )

    if not keep_id and "id" in to_return.attrs:
        del to_return.attrs["id"]

    provenance(
        to_return,
        arr,
        {
            "what": "Normalize axis or axes",
            "by": "normalize_dim",
            "dims": dims,
        },
    )

    return to_return


@lift_dataarray_to_generic
def normalize_dim(arr: DataType, dims: Container[str] | str):
    """Normalizes the intensity so that all values along axes specfied by "dims" are the same.

    The function normalizes so that the average value of cells in the output is 1.

    Args:
        arr: Input data which should be normalized
        dim_o=r_dims: The set of dimensions which should be normalized
        keep_id: Whether to reset the data's id after normalization

    Returns:
        The normalized data.
    """
    dims = [dims] if isinstance(dims, str) else dims

    summed_arr = arr.fillna(arr.mean()).sum([d for d in arr.dims if d not in dims])
    normalized_arr = arr / (summed_arr / np.prod(summed_arr.shape))

    to_return = xr.DataArray(normalized_arr.values, arr.coords, arr.dims)

    return to_return


@update_provenance("Normalize total spectrum intensity")
def normalize_total(data: DataType):
    """Normalizes data so that the total intensity is 1000000 (a bit arbitrary)."""
    data = normalize_to_spectrum(data)

    return data / (data.sum(data.dims) / 1000000)


def dim_normalizer(dim_name):
    """Safe partial application of dimension normalization."""

    def normalize(arr: xr.DataArray):
        if dim_name not in arr.dims:
            return arr
        return normalize_dim(arr, dim_name)

    return normalize


def transform_dataarray_axis(
    f,
    old_axis_name: str,
    new_axis_name: str,
    new_axis,
    dataset: xr.DataArray,
    prep_name,
    transform_spectra=None,
    remove_old=True,
):
    """Applies a function onto a DataArray axis."""
    ds = dataset.copy()
    if transform_spectra is None:
        # transform *all* DataArrays in the dataset that have old_axis_name in their dimensions
        transform_spectra = {
            k: v for k, v in ds.data_vars.items() if old_axis_name in v.dims
        }

    ds.coords[new_axis_name] = new_axis

    new_dataarrays = []
    for name in transform_spectra.keys():
        dr = ds[name]

        old_axis = dr.dims.index(old_axis_name)
        shape = list(dr.sizes.values())
        shape[old_axis] = len(new_axis)
        new_dims = list(dr.dims)
        new_dims[old_axis] = new_axis_name

        g = functools.partial(f, axis=old_axis)
        output = geometric_transform(
            dr.values, g, output_shape=shape, output="f", order=1
        )

        new_coords = dict(dr.coords)
        new_coords.pop(old_axis_name)

        new_dataarray = xr.DataArray(
            output,
            coords=new_coords,
            dims=new_dims,
            attrs=dr.attrs.copy(),
            name=prep_name(dr.name),
        )
        new_dataarrays.append(new_dataarray)
        if "id" in new_dataarray.attrs:
            del new_dataarray.attrs["id"]

        if remove_old:
            del ds[name]
        else:
            assert prep_name(name) != name and "You must make sure names don't collide"

    new_ds = xr.merge([ds, *new_dataarrays])

    new_ds.attrs.update(ds.attrs.copy())

    if "id" in new_ds:
        del new_ds.attrs["id"]

    provenance(
        new_ds,
        dataset,
        {
            "what": "Transformed a Dataset coordinate axis",
            "by": "transform_dataarray_axis",
            "old_axis": old_axis_name,
            "new_axis": new_axis_name,
            "transformed_vars": list(transform_spectra.keys()),
        },
    )

    return new_ds



"""Utilities related to statistical bootstraps.

It can sometimes be difficult to assess when bootstraps are appropriate, 
so make sure to consider this before you just stick a bootstrap around 
your code and stuff the resultant error bar into your papers.

This is most useful on data coming from ToF experiments, where individual electron
arrivals are counted, but even here you must be aware of tricky aspects of
the experiment: ToF-ARPES analyzers are not perfect, their efficiency can vary dramatically
across the detector due to MCP burn-in, and electron aberration and focusing
must be considered.
"""

import copy
from dataclasses import dataclass

import functools
import random

import scipy.stats
from typing import Any, Callable, Dict, List, Optional, Set, Union
import numpy as np
from tqdm import tqdm_notebook

import xarray as xr
from arpes.analysis.sarpes import to_intensity_polarization
from arpes.provenance import update_provenance
from arpes.typing import DataType
from arpes.utilities import lift_dataarray_to_generic
from arpes.utilities.normalize_datatype import normalize_to_spectrum
from arpes.utilities.region import normalize_region

__all__ = (
    "bootstrap",
    "estimate_prior_adjustment",
    "resample_true_counts",
    "bootstrap_counts",
    "bootstrap_intensity_polarization",
    "Normal",
    "propagate_errors",
)


@update_provenance("Estimate prior")
def estimate_prior_adjustment(
    data: DataType, region: Union[Dict[str, Any], str] = None
) -> float:
    r"""Estimates the parameters of a distribution generating the intensity histogram of pixels in a spectrum.

    In a perfectly linear, single-electron
    single-count detector, this would be a poisson distribution with
    \lambda=mean(counts) over the window. Despite this, we can estimate \lambda
    phenomenologically and verify that a Poisson distribution provides a good
    prior for the data, allowing us to perform statistical bootstrapping.

    You should use this with a spectrum that has uniform intensity, i.e. with a
    copper reference or similar.

    Args:
        data: The input spectrum.
        region: The region which should be used for the estimate.

    Returns:
        sigma / mu, the adjustment factor for the Poisson distribution
    """
    data = normalize_to_spectrum(data)

    if region is None:
        region = "copper_prior"

    region = normalize_region(region)

    if "cycle" in data.dims:
        data = data.sum("cycle")

    data = data.S.zero_spectrometer_edges().S.region_sel(region)
    values = data.values.ravel()
    values = values[np.where(values)]
    return np.std(values) / np.mean(values)


@update_provenance("Resample cycle dimension")
@lift_dataarray_to_generic
def resample_cycle(data: xr.DataArray, **kwargs) -> xr.DataArray:
    """Perform a non-parametric bootstrap using a cycle coordinate for statistically independent observations.

    Args:
        data: The input data.
        kwargs: Unused

    Returns:
        Resampled data with selections from the cycle axis.
    """
    n_cycles = len(data.cycle)
    which = [random.randint(0, n_cycles - 1) for _ in range(n_cycles)]

    resampled = data.isel(cycle=which).sum("cycle", keep_attrs=True)

    if "id" in resampled.attrs:
        del resampled.attrs["id"]

    return resampled


@update_provenance("Resample with prior adjustment")
@lift_dataarray_to_generic
def resample(data: xr.DataArray, prior_adjustment=1, **kwargs):
    resampled = xr.DataArray(
        np.random.poisson(lam=data.values * prior_adjustment, size=data.values.shape),
        coords=data.coords,
        dims=data.dims,
        attrs=data.attrs,
    )

    if "id" in resampled.attrs:
        del resampled.attrs["id"]

    return resampled


@update_provenance("Resample electron-counted data")
@lift_dataarray_to_generic
def resample_true_counts(data: xr.DataArray) -> xr.DataArray:
    """Resamples histogrammed data where each count represents an actual electron.

    Args:
        data: Input data representing actual electron counts from a time of flight
              system or delay line.

    Returns:
        Poisson resampled data.
    """
    resampled = xr.DataArray(
        np.random.poisson(lam=data.values, size=data.values.shape),
        coords=data.coords,
        dims=data.dims,
        attrs=data.attrs,
    )

    if "id" in resampled.attrs:
        del resampled.attrs["id"]

    return resampled


@update_provenance("Bootstrap true electron counts")
@lift_dataarray_to_generic
def bootstrap_counts(data: DataType, N=1000, name=None) -> xr.Dataset:
    """Performs a parametric bootstrap assuming recorded data are electron counts.

    Parametric bootstrap for the number of counts in each detector channel for a
    time of flight/DLD detector, where each count represents an actual particle.

    This function also introspects the data passed to determine whether there is a
    spin degree of freedom, and will bootstrap appropriately.

    Currently we build all the samples at once instead of using a rolling algorithm.

    Arguments:
        data: The input spectrum.
        N: The number of samples to draw.
        name: The name of the subarray which represents counts to resample. E.g. "up_spectrum"

    Returns:
        A `xr.Dataset` which has the mean and standard error for the resampled named array.
    """
    assert data.name is not None or name is not None
    name = data.name if data.name is not None else name

    desc_fragment = " {}".format(name)

    resampled_sets = []
    for _ in tqdm_notebook(range(N), desc="Resampling{}...".format(desc_fragment)):
        resampled_sets.append(resample_true_counts(data))

    resampled_arr = np.stack([s.values for s in resampled_sets], axis=0)
    std = np.std(resampled_arr, axis=0)
    std = xr.DataArray(std, data.coords, data.dims)
    mean = np.mean(resampled_arr, axis=0)
    mean = xr.DataArray(mean, data.coords, data.dims)

    data_vars = {}
    data_vars[name] = mean
    data_vars[name + "_std"] = std

    return xr.Dataset(data_vars=data_vars, coords=data.coords, attrs=data.attrs.copy())


class Distribution:
    DEFAULT_N_SAMPLES = 1000

    def draw_samples(self, n_samples=DEFAULT_N_SAMPLES):
        """Draws samples from this distribution."""
        raise NotImplementedError


@dataclass
class Normal(Distribution):
    """Represents a Gaussian distribution.

    Attributes:
        center: The center/mu parameter for the distribution.
        stderr: The standard error for the distribution.
    """

    center: float
    stderr: float

    def draw_samples(self, n_samples=Distribution.DEFAULT_N_SAMPLES):
        """Draws samples from this distribution."""
        return scipy.stats.norm.rvs(self.center, scale=self.stderr, size=n_samples)

    @classmethod
    def from_param(cls, model_param):
        """Generates a Normal from an `lmfit.Parameter`."""
        return cls(center=model_param.value, stderr=model_param.stderr)


def propagate_errors(f) -> Callable:
    """A decorator which provides transparent propagation of statistical errors.

    The way that this is accomodated is that the inner function is turned into one which
    operates over distributions. Errors are calculated empirically by sampling
    over trials drawn from these distributions.

    CAVEAT EMPTOR: Arguments are assumed to be uncorrelated.

    Args:
        f: The inner function

    Returns:
        The wrapped function.
    """

    @functools.wraps(f)
    def operates_on_distributions(*args, **kwargs):
        exclude = set(
            [i for i, arg in enumerate(args) if not isinstance(arg, Distribution)]
            + [k for k, arg in kwargs.items() if not isinstance(arg, Distribution)]
        )

        if len(exclude) == len(args) + len(kwargs):
            # short circuit if no bootstrapping is required to be nice to the user
            return f(*args, **kwargs)

        vec_f = np.vectorize(f, excluded=exclude)
        res = vec_f(
            *[a.draw_samples() if isinstance(a, Distribution) else a for a in args],
            **{
                k: v.draw_samples() if isinstance(v, Distribution) else v
                for k, v in kwargs.items()
            }
        )

        try:
            print(scipy.stats.describe(res))
        except:
            pass

        return res

    return operates_on_distributions


@update_provenance("Bootstrap spin detector polarization and intensity")
def bootstrap_intensity_polarization(data: xr.Dataset, N: int = 100) -> xr.Dataset:
    """Builds an estimate of the intensity and polarization from spin-data.

    Uses the parametric bootstrap to get uncertainties on the intensity and polarization of ToF-SARPES data.

    Args:
        data: Input spectrum for resampling.
        N: The number of samples to draw.

    Returns:
        Resampled data after conversion to intensity and polarization.
    """
    bootstrapped_polarization = bootstrap(to_intensity_polarization)
    return bootstrapped_polarization(data, N=N)


def bootstrap(
    fn: Callable,
    skip: Optional[Union[Set[int], List[int]]] = None,
    resample_method: Optional[str] = None,
) -> Callable:
    """Produces function which performs a bootstrap of an arbitrary function by sampling.

    This is a functor which takes a function operating on plain data and produces one which
    internally bootstraps over counts on the input data.

    Args:
        fn: The function to be bootstrapped.
        skip: Which arguments to leave alone. Defaults to None.
        resample_method: How the resampling should be performed. See `resample` and `resample_cycle`. Defaults to None.

    Returns:
        A function which vectorizes the ouptut of the input function `fn` over samples.
    """
    if skip is None:
        skip = []

    skip = set(skip)

    if resample_method is None:
        resample_fn = resample
    elif resample_method == "cycle":
        resample_fn = resample_cycle

    def bootstrapped(*args, N=20, prior_adjustment=1, **kwargs):
        # examine args to determine which to resample
        resample_indices = [
            i
            for i, arg in enumerate(args)
            if isinstance(arg, (xr.DataArray, xr.Dataset)) and i not in skip
        ]
        data_is_arraylike = False

        runs = []

        def get_label(i):
            if isinstance(args[i], xr.Dataset):
                return "xr.Dataset: [{}]".format(", ".join(args[i].data_vars.keys()))
            if args[i].name:
                return args[i].name

            try:
                return args[i].attrs["id"]
            except KeyError:
                return "Label-less DataArray"

        print(
            "Resampling args: {}".format(
                ",".join([get_label(i) for i in resample_indices])
            )
        )

        # examine kwargs to determine which to resample
        resample_kwargs = [
            k
            for k, v in kwargs.items()
            if isinstance(v, xr.DataArray) and k not in skip
        ]
        print("Resampling kwargs: {}".format(",".join(resample_kwargs)))

        print(
            "Fair warning 1: Make sure you understand whether it is appropriate to resample your data."
        )
        print(
            "Fair warning 2: Ensure that the data to resample is in a DataArray and not a Dataset"
        )

        for _ in tqdm_notebook(range(N), desc="Resampling..."):
            new_args = list(args)
            new_kwargs = copy.copy(kwargs)
            for i in resample_indices:
                new_args[i] = resample_fn(args[i], prior_adjustment=prior_adjustment)
            for k in resample_kwargs:
                new_kwargs[k] = resample_fn(
                    kwargs[k], prior_adjustment=prior_adjustment
                )

            run = fn(*new_args, **new_kwargs)
            if isinstance(run, (xr.DataArray, xr.Dataset)):
                data_is_arraylike = True
            runs.append(run)

        if data_is_arraylike:
            for i, run in enumerate(runs):
                run = run.assign_coords(bootstrap=i)

            return xr.concat(runs, dim="bootstrap")

        return runs

    return functools.wraps(fn)(bootstrapped)



"""Provides background estimation approaches."""
import xarray as xr
import numpy as np
from scipy.spatial import ConvexHull
from scipy.interpolate import interp1d

__all__ = (
    "calculate_background_hull",
    "remove_background_hull",
)


def calculate_background_hull(arr, breakpoints=None):
    """Calculates a background using the convex hull of the data (viewing the intensity as a Z axis)."""
    if breakpoints:
        breakpoints = [None] + breakpoints + [None]
        dim = arr.dims[0]
        processed = []
        for blow, bhigh in zip(breakpoints, breakpoints[1:]):
            processed.append(
                calculate_background_hull(arr.sel(**dict([[dim, slice(blow, bhigh)]])))
            )
        return xr.concat(processed, dim)

    points = np.stack(arr.G.to_arrays(), axis=1)
    hull = ConvexHull(points)

    vertices = np.array(hull.vertices)
    index_of_zero = np.argwhere(vertices == 0)[0][0]
    vertices = np.roll(vertices, -index_of_zero)
    xis = vertices[: np.argwhere(vertices == len(arr) - 1)[0][0]]
    xis = list(xis) + [len(arr) - 1]

    support = points[xis]
    bkg = interp1d(support[:, 0], support[:, 1], fill_value="extrapolate")(points[:, 0])
    return arr.S.with_values(interp1d(support[:, 0], support[:, 1])(points[:, 0]))


def remove_background_hull(data, *args, **kwargs):
    """Removes a background according to `calculate_background_hull`."""
    return data - calculate_background_hull(data, *args, **kwargs)
